<?php

if (!defined('MEDIAWIKI')) die();

/**
 * The SFFProperty class represents a SMW Property.
 *
 * @author Henri Gerrits
 */
class SFFProperty {

    // instance variables - see doc on getters/setters
    // mSavedXxx are retrieved from the existing article
    // mXxx are the same, except where overridden by the request parameters, i.e., the new values
    private $mName;
    private $mType;
    private $mSavedType;
    private $mRange;
    private $mSavedRange;
    private $mDomain;
    private $mSavedDomain;
    private $mInverseOf;
    private $mSavedInverseOf;
    private $mSymmetric;
    private $mSavedSymmetric;
    private $mLabel;
    private $mSavedLabel;
    private $mComment;
    private $mSavedComment;
    private $mDefaultForm;
    private $mSavedDefaultForm;
    private $mAllowedValues;
    private $mSavedAllowedValues;

    // transient variables
    private $mTitle;

    // patterns for properties
    private static $sTypeLinkPattern;
    private static $sRangeParserFunctionPattern;
    private static $sDomainParserFunctionPattern;
    private static $sInverseOfParserFunctionPattern;
    private static $sSymmetricLinkPattern;
    private static $sLabelLinkPattern;
    private static $sCommentLinkPattern;
    private static $sDefaultFormLinkPattern;
    private static $sAllowsValuesParserFunctionPattern;

    /**
     * Constructs a new SFFProperty. The new values for the meta properties are taken from the
     * request parameters, if present. The saved values of the meta properties are loaded from the
     * article.
     * @param WebRequest $request optional web request containing new values for the meta properties
     */
    function __construct(WebRequest $request = null) {
        self::buildPropertyLinkPatterns();
        if (isset($request)) {
            $this->setName($request->getVal('name'));
            if ($this->hasArticle()) {
                // get the original values from the wikitext
                $this->loadSavedProperties();
            }
            // set 'new' values from the request parameters
            if ($request->getCheck('type')) {
                $this->setType($request->getVal('type'));
            }
            if ($request->getCheck('range')) {
                $this->setRange(self::getArrayParameter($request, 'range'));
            }
            if ($request->getCheck('domain')) {
                $this->setDomain(self::getArrayParameter($request, 'domain'));
            }
            if ($request->getCheck('inverseof')) {
                $this->setInverseOf(self::getArrayParameter($request, 'inverseof'));
            }
            if ($request->getCheck('symmetric')) {
                $this->setSymmetric($request->getVal('symmetric'));
            }
            if ($request->getCheck('label')) {
                $this->setLabel($request->getVal('label'));
            }
            if ($request->getCheck('comment')) {
                $this->setComment($request->getVal('comment'));
            }
            if ($request->getCheck('defaultform')) {
                $this->setDefaultForm($request->getVal('defaultform'));
            }
            if ($request->getCheck('allowedvalues')) {
                $listSeparator = SFFGlobals::getListSeparator();
                $allowedValuesString = $request->getVal('allowedvalues');
                // replace escaped list separator temporarily with ASCII bell (which the user could not have entered)
                $allowedValuesString = str_replace("\\$listSeparator", "\007", $allowedValuesString);
                // split among list separator, possibly surrounded by whitespace
                $allowedValuesArray = preg_split("/\s*$listSeparator\s*/", $allowedValuesString, -1, PREG_SPLIT_NO_EMPTY);
                foreach($allowedValuesArray as &$value) {
                    // now replace ASCII bell with list separator again
                    $value = str_replace("\007", $listSeparator, $value);
                }
                $this->setAllowedValues($allowedValuesArray);
            }
        }
    }

    private static function getArrayParameter(WebRequest $request, $parameterName) {
        $listSeparator = SFFGlobals::getListSeparator();
        $parameterString = $request->getVal($parameterName);
        // split among list separator, possibly surrounded by whitespace
        return preg_split("/\s*$listSeparator\s*/", $parameterString, -1, PREG_SPLIT_NO_EMPTY);
    }

    /**
     * Builds the meta property link patterns used for searching article text.
     */
    private static function buildPropertyLinkPatterns() {
        if (!isset(self::$sTypeLinkPattern)) {
            // @todo document
            // @todo internationalize
            // @todo use tags
            // @todo cater for alternative text (see SMW_ParserExtensions)
            // @todo cater for multiple assignment
            self::$sTypeLinkPattern = '\[\[\s*has(?:\s|_)+type\s*::\s*(.*?)\s*\]\]';
            self::$sRangeParserFunctionPattern = '\{\{\s*#\s*range\s*:\s*(.*?)\s*\}\}';
            self::$sDomainParserFunctionPattern = '\{\{\s*#\s*domain\s*:\s*(.*?)\s*\}\}';
            self::$sInverseOfParserFunctionPattern = '\{\{\s*#\s*inverse(?:\s|_)+of\s*:\s*(.*?)\s*\}\}';
            self::$sSymmetricLinkPattern = '\[\[\s*is(?:\s|_)+symmetric\s*::\s*(.*?)\s*\]\]';
            self::$sLabelLinkPattern = '\[\[\s*has(?:\s|_)+label\s*::\s*(.*?)\s*\]\]';
            self::$sCommentLinkPattern = '\[\[\s*has(?:\s|_)+comment\s*::\s*(.*?)\s*\]\]';
            self::$sDefaultFormLinkPattern = '\[\[\s*has(?:\s|_)+default(?:\s|_)+form\s*::\s*form:\s*(.*?)\s*\]\]';
            self::$sAllowsValuesParserFunctionPattern = '\{\{\s*#\s*allows\s+values\s*:\s*(.*?)\s*\}\}';
        }
    }

    /**
     * Returns the PCRE pattern for type links, including delimiters and modifiers. The value of the
     * link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for type links, including delimiters and modifiers
     */
    static function getTypeLinkPattern() {
        return '/' . self::getRawTypeLinkPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for type links, excluding delimiters or modifiers. The value of the
     * link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for type links, excluding delimiters or modifiers
     */
    static function getRawTypeLinkPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sTypeLinkPattern;
    }

    /**
     * Returns the PCRE pattern for has range parser function calls, including delimiters and
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, including delimiters
     * and modifiers
     */
    static function getRangeParserFunctionPattern() {
        return '/' . self::getRawRangeParserFunctionPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for has range parser function calls, excluding delimiters or
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, excluding delimiters
     * or modifiers
     */
    static function getRawRangeParserFunctionPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sRangeParserFunctionPattern;
    }

    /**
     * Returns the PCRE pattern for has domain parser function calls, including delimiters and
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, including delimiters
     * and modifiers
     */
    static function getDomainParserFunctionPattern() {
        return '/' . self::getRawDomainParserFunctionPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for has domain parser function calls, excluding delimiters or
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, excluding delimiters
     * or modifiers
     */
    static function getRawDomainParserFunctionPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sDomainParserFunctionPattern;
    }

    /**
     * Returns the PCRE pattern for is inverse of parser function calls, including delimiters and
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, including delimiters
     * and modifiers
     */
    static function getInverseOfParserFunctionPattern() {
        return '/' . self::getRawInverseOfParserFunctionPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for is inverse of parser function calls, excluding delimiters or
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, excluding delimiters
     * or modifiers
     */
    static function getRawInverseOfParserFunctionPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sInverseOfParserFunctionPattern;
    }

     /**
     * Returns the PCRE pattern for symmetric links, including delimiters and modifiers. The
     * value of the link will be available as subpattern 1 as described in the documentation for the
     * PHP preg_match() function.
     * @return String the PCRE pattern for default form links, including delimiters and modifiers
     */
    static function getSymmetricLinkPattern() {
        return '/' . self::getRawSymmetricLinkPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for symmetric links, excluding delimiters or modifiers. The value
     * of the link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for default form links, excluding delimiters or modifiers
     */
    static function getRawSymmetricLinkPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sSymmetricLinkPattern;
    }

     /**
     * Returns the PCRE pattern for label links, including delimiters and modifiers. The
     * value of the link will be available as subpattern 1 as described in the documentation for the
     * PHP preg_match() function.
     * @return String the PCRE pattern for default form links, including delimiters and modifiers
     */
    static function getLabelLinkPattern() {
        return '/' . self::getRawLabelLinkPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for label links, excluding delimiters or modifiers. The value
     * of the link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for default form links, excluding delimiters or modifiers
     */
    static function getRawLabelLinkPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sLabelLinkPattern;
    }

     /**
     * Returns the PCRE pattern for comment links, including delimiters and modifiers. The
     * value of the link will be available as subpattern 1 as described in the documentation for the
     * PHP preg_match() function.
     * @return String the PCRE pattern for default form links, including delimiters and modifiers
     */
    static function getCommentLinkPattern() {
        return '/' . self::getRawCommentLinkPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for comment links, excluding delimiters or modifiers. The value
     * of the link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for default form links, excluding delimiters or modifiers
     */
    static function getRawCommentLinkPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sCommentLinkPattern;
    }

     /**
     * Returns the PCRE pattern for default form links, including delimiters and modifiers. The
     * value of the link will be available as subpattern 1 as described in the documentation for the
     * PHP preg_match() function.
     * @return String the PCRE pattern for default form links, including delimiters and modifiers
     */
    static function getDefaultFormLinkPattern() {
        return '/' . self::getRawDefaultFormLinkPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for default form links, excluding delimiters or modifiers. The value
     * of the link will be available as subpattern 1 as described in the documentation for the PHP
     * preg_match() function.
     * @return String the PCRE pattern for default form links, excluding delimiters or modifiers
     */
    static function getRawDefaultFormLinkPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sDefaultFormLinkPattern;
    }

    /**
     * Returns the PCRE pattern for allows values parser function calls, including delimiters and
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, including delimiters
     * and modifiers
     */
    static function getAllowsValuesParserFunctionPattern() {
        return '/' . self::getRawAllowsValuesParserFunctionPattern() . '/i';
    }

    /**
     * Returns the PCRE pattern for allows values parser function calls, excluding delimiters or
     * modifiers. The unparsed parameter list of the function call will be available as subpattern 1
     * as described in the documentation for the PHP preg_match() function.
     * @return String the PCRE pattern for allows values parser function calls, excluding delimiters
     * or modifiers
     */
    static function getRawAllowsValuesParserFunctionPattern() {
        self::buildPropertyLinkPatterns();
        return self::$sAllowsValuesParserFunctionPattern;
    }

    /**
     * Reads the meta property values from the wikitext using pattern matching. Only the first
     * occurrences are used.
     */
    private function loadSavedProperties() {
        // @todo use named patterns
        // this assumes that $mName has already been set
        $wikiText = $this->getSavedWikiText();
        $typeMatches = preg_match(self::getTypeLinkPattern(), $wikiText, $matches);
        if ($typeMatches) {
            $this->mSavedType = $matches[1];
            $this->setType($this->mSavedType);
        }
        $rangeMatches = preg_match(self::getRangeParserFunctionPattern(), $wikiText, $matches);
        if ($rangeMatches) {
            // get the values part
            $rangeList = $matches[1];
            // split this into an array along the pipe symbol, absorbing any surrounding whitespace as well
            $this->mSavedRange = preg_split('/\s*\|\s*/', $rangeList);
            $this->setRange($this->mSavedRange);
        }
        $domainMatches = preg_match(self::getDomainParserFunctionPattern(), $wikiText, $matches);
        if ($domainMatches) {
            // get the values part
            $domainList = $matches[1];
            // split this into an array along the pipe symbol, absorbing any surrounding whitespace as well
            $this->mSavedDomain = preg_split('/\s*\|\s*/', $domainList);
            $this->setDomain($this->mSavedDomain);
        }
        $inverseOfMatches = preg_match(self::getInverseOfParserFunctionPattern(), $wikiText, $matches);
        if ($inverseOfMatches) {
            // get the values part
            $inverseOfList = $matches[1];
            // split this into an array along the pipe symbol, absorbing any surrounding whitespace as well
            $this->mSavedInverseOf = preg_split('/\s*\|\s*/', $inverseOfList);
            $this->setInverseOf($this->mSavedInverseOf);
        }
        $symmetricMatches = preg_match(self::getSymmetricLinkPattern(), $wikiText, $matches);
        if ($symmetricMatches) {
            $this->mSavedSymmetric = $matches[1];
            $this->setSymmetric($this->mSavedSymmetric);
        }
        $labelMatches = preg_match(self::getLabelLinkPattern(), $wikiText, $matches);
        if ($labelMatches) {
            $this->mSavedLabel = $matches[1];
            $this->setLabel($this->mSavedLabel);
        }
        $commentMatches = preg_match(self::getCommentLinkPattern(), $wikiText, $matches);
        if ($commentMatches) {
            $this->mSavedComment = $matches[1];
            $this->setComment($this->mSavedComment);
        }
        $defaultformMatches = preg_match(self::getDefaultFormLinkPattern(), $wikiText, $matches);
        if ($defaultformMatches) {
            $this->mSavedDefaultForm = $matches[1];
            $this->setDefaultForm($this->mSavedDefaultForm);
        }
        $allowsValuesMatches = preg_match(self::getAllowsValuesParserFunctionPattern(), $wikiText, $matches);
        if ($allowsValuesMatches) {
            // get the values part
            $allowedValuesList = $matches[1];
            // split this into an array along the pipe symbol, absorbing any surrounding whitespace as well
            $this->mSavedAllowedValues = preg_split('/\s*\|\s*/', $allowedValuesList);
            $this->setAllowedValues($this->mSavedAllowedValues);
        }
    }

    /**
     * Returns the name of this SFFProperty.
     * @return String the name of this SFFProperty
     */
    function getName() {
        return $this->mName;
    }

    /**
     * Sets the name of this SFFProperty. Excess whitespace is removed, and underscores are replaced
     * by spaces.
     * @param String $name the name of this SFFProperty
     */
    function setName($name = '') {
        // reduce any whitespace or underscores between words to a single space
        $name = preg_replace('/(\s|_)+/', ' ', $name);
        $this->mName = trim($name);
        // cache title object
        $this->mTitle = Title::makeTitleSafe(SMW_NS_PROPERTY, $this->getName());
    }

    /**
     * Returns the type (more precisely, the type's name) of this property. In wikitext, this is set
     * by the 'Has type' meta property.
     * @return String the type (name) of this property
     */
    function getType() {
        return $this->mType;
    }

    /**
     * Sets the type (more precisely, the type's name) of this property. In wikitext, this is set
     * by the 'Has type' meta property.
     * @param String type the new type (name) for this property
     */
    function setType($type = '') {
        $this->mType = $type;
    }

    /**
     * Returns the type of this property as an annotated wikitext link. It has the form
     * [[Has type::<type name>]] in English.
     * @return String an annotated wikitext link to the type of this property
     */
    function getTypeLink() {
        return "[[{$this->getTypeAnnotationTag()}::{$this->getType()}]]";
    }

    /**
     * Returns true if a type has now been defined for this property, where no type was defined in
     * the property's saved article.
     * @return Boolean true if the type has now been added, false otherwise
     */
    function isTypeAdded() {
        return $this->mType && !$this->mSavedType;
    }

    /**
     * Returns true if no type is defined for this property, where there was a type defined in the
     * property's saved article.
     * @return Boolean true if the type has now been removed, false otherwise
     */
    function isTypeRemoved() {
        return !$this->mType && $this->mSavedType;
    }

    /**
     * Returns true if a type is defined for this property, where there was a different type defined
     * in the property's saved article.
     * @return Boolean true if the type has been changed, false otherwise
     */
    function isTypeChanged() {
        // no overlap with isTypeAdded or isTypeRemoved
        return $this->mType && $this->mSavedType && $this->mType != $this->mSavedType;
    }

    /**
     * Returns an array of the range of this property. In wikitext, this is set by the 'Has range'
     * parser function call. The array is not guaranteed to contain only String values.
     * @return array the range of this property
     */
    function getRange() {
        return $this->mRange;
    }

    /**
     * Sets the range of this property as an array of category names. In wikitext, this is set by
     * the 'Has range' parser function call. Duplicate items (case sensitive) will be removed. The
     * argument's elements are not verified to be of type String.
     * @param array range the (possibly null or empty) array of category names
     */
    function setRange(array $range = null) {
        $this->mRange = array_unique($range);
    }

    /**
     * Returns the range of this property as a wikitext parser function call. It has the form
     * {{#Has range:<category1>|<category2>|<...>}} in English.
     * @return String a wikitext parser function call for the range of this property
     */
    function getRangeFunctionCall() {
        $tag = $this->getRangeCallTag();
        $values = $this->getRange();
        return '{{' . $tag . ':' . implode('|', $values) . '}}'; // clumsy, but cannot escape '{'
     }

    /**
     * Returns true if a range has now been defined for this property, where no range was defined in
     * the property's saved article.
     * @return Boolean true if the range has now been added, false otherwise
     */
    function isRangeAdded() {
        return (count($this->mRange) > 0 && count($this->mSavedRange) == 0);
    }

    /**
     * Returns true if no range is defined for this property, where there was a range defined in the
     * property's saved article.
     * @return Boolean true if the range has now been removed, false otherwise
     */
    function isRangeRemoved() {
        return (count($this->mRange) == 0 && count($this->mSavedRange) > 0);
    }

    /**
     * Returns true if a range is defined for this property, where there was a different range
     * defined in the property's saved article.
     * @return Boolean true if the range has been changed, false otherwise
     */
    function isRangeChanged() {
        // no overlap with isRangeAdded or isRangeRemoved
        return (count($this->mRange) > 0 && count($this->mSavedRange) > 0 && $this->mRange != $this->mSavedRange);
    }

    /**
     * Returns an array of the domain of this property. In wikitext, this is set by the 'Has domain'
     * parser function call. The array is not guaranteed to contain only String values.
     * @return array the domain of this property
     */
    function getDomain() {
        return $this->mDomain;
    }

    /**
     * Sets the domain of this property as an array of category names. In wikitext, this is set by
     * the 'Has domain' parser function call. Duplicate items (case sensitive) will be removed. The
     * argument's elements are not verified to be of type String.
     * @param array domain the (possibly null or empty) array of category names
     */
    function setDomain(array $domain = null) {
        $this->mDomain = array_unique($domain);
    }

    /**
     * Returns the domain of this property as a wikitext parser function call. It has the form
     * {{#Has domain:<category1>|<category2>|<...>}} in English.
     * @return String a wikitext parser function call for the domain of this property
     */
    function getDomainFunctionCall() {
        $tag = $this->getDomainCallTag();
        $values = $this->getDomain();
        return '{{' . $tag . ':' . implode('|', $values) . '}}'; // clumsy, but cannot escape '{'
     }

    /**
     * Returns true if a domain has now been defined for this property, where no domain was defined
     * in the property's saved article.
     * @return Boolean true if the domain has now been added, false otherwise
     */
    function isDomainAdded() {
        return (count($this->mDomain) > 0 && count($this->mSavedDomain) == 0);
    }

    /**
     * Returns true if no domain is defined for this property, where there was a domain defined in
     * the property's saved article.
     * @return Boolean true if the domain has now been removed, false otherwise
     */
    function isDomainRemoved() {
        return (count($this->mDomain) == 0 && count($this->mSavedDomain) > 0);
    }

    /**
     * Returns true if a domain is defined for this property, where there was a different domain
     * defined in the property's saved article.
     * @return Boolean true if the domain has been changed, false otherwise
     */
    function isDomainChanged() {
        // no overlap with isDomainAdded or isDomainRemoved
        return (count($this->mDomain) > 0 && count($this->mSavedDomain) > 0 && $this->mDomain != $this->mSavedDomain);
    }

    /**
     * Returns an array of properties of which this property is the inverse. In wikitext, this is
     * set by the 'Is inverse of' parser function call. The array is not guaranteed to contain only
     * String values.
     * @return array the properties of which this property is the inverse
     */
    function getInverseOf() {
        return $this->mInverseOf;
    }

    /**
     * Sets the properties of which this property is the inverse as an array of property names. In
     * wikitext, this is set by the 'Is inverse of' parser function call. Duplicate items (case
     * sensitive) will be removed. The argument's elements are not verified to be of type String.
     * @param array inverseOf the (possibly null or empty) array of property names
     */
    function setInverseOf(array $inverseOf = null) {
        $this->mInverseOf = array_unique($inverseOf);
    }

    /**
     * Returns the properties of which this property is the inverse as a wikitext parser function
     * call. It has the form {{#Is inverse of:<property1>|<property2>|<...>}} in English.
     * @return String a wikitext parser function call for the properties of which this property is
     * the inverse
     */
    function getInverseOfFunctionCall() {
        $tag = $this->getInverseOfCallTag();
        $values = $this->getInverseOf();
        return '{{' . $tag . ':' . implode('|', $values) . '}}'; // clumsy, but cannot escape '{'
     }

    /**
     * Returns true if inverseOf properties have now been defined for this property, where no
     * inverseOf properties were defined in the property's saved article.
     * @return Boolean true if inverseOf properties have now been added, false otherwise
     */
    function isInverseOfAdded() {
        return (count($this->mInverseOf) > 0 && count($this->mSavedInverseOf) == 0);
    }

    /**
     * Returns true if no inverseOf properties are defined for this property, where there were
     * inverseOf properties defined in the property's saved article.
     * @return Boolean true if inverseOf properties have now been removed, false otherwise
     */
    function isInverseOfRemoved() {
        return (count($this->mInverseOf) == 0 && count($this->mSavedInverseOf) > 0);
    }

    /**
     * Returns true if a inverseOf properties are defined for this property, where there were
     * different inverseOf properties defined in the property's saved article.
     * @return Boolean true if inverseOf properties have been changed, false otherwise
     */
    function isInverseOfChanged() {
        // no overlap with isInverseOfAdded or isInverseOfRemoved
        return (count($this->mInverseOf) > 0 && count($this->mSavedInverseOf) > 0 && $this->mInverseOf != $this->mSavedInverseOf);
    }

    /**
     * Returns whether this property is symmetric. In wikitext, this is set by the 'Is symmetric'
     * meta property.
     * @return String 'true' if this property is symmetric, 'false' otherwise
     */
    function getSymmetric() {
        return $this->mSymmetric;
    }

    /**
     * Sets whether this property is symmetric. In wikitext, this is set by the 'Is symmetric' meta
     * property.
     * @param String 'true' if this property is symmetric, 'false' otherwise
     */
    function setSymmetric($symmetric = '') {
        $this->mSymmetric = $symmetric;
    }

    /**
     * Returns whether this property is symmetric as an annotated wikitext link. It has the form
     * [[Is symmetric::<true or false>]] in English.
     * @return String an annotated wikitext link for the Is symmetric meta property
     */
    function getSymmetricLink() {
        $tag = $this->getSymmetricAnnotationTag();
        return "[[$tag::$this->mSymmetric]]";
    }

    /**
     * Returns true if the Is symmetric meta property has now been defined for this property,
     * where no Is symmetric meta property was defined in the property's saved article.
     * @return Boolean true if an Is symmetric meta property has now been added, false otherwise
     */
     function isSymmetricAdded() {
        return $this->mSymmetric && !$this->mSavedSymmetric;
    }

    /**
     * Returns true if no Is symmetric meta property is defined for this property, where the Is
     * symmetric meta property was defined in the property's saved article.
     * @return Boolean true if the Is symmetric meta property has now been removed, false otherwise
     */
    function isSymmetricRemoved() {
        return !$this->mSymmetric && $this->mSavedSymmetric;
    }

    /**
     * Returns true if the Is symmetric is defined for this property, where there was a different Is
     * symmetric meta property defined in the property's saved article.
     * @return Boolean true if the Is symmetric meta property has been changed, false otherwise
     */
    function isSymmetricChanged() {
        // no overlap with isSymmetricAdded or isSymmetricRemoved
        return $this->mSymmetric && $this->mSavedSymmetric && $this->mSymmetric != $this->mSavedSymmetric;
    }

    /**
     * Returns the label of this property. In wikitext, this is set by the 'Has label' meta
     * property.
     * @return String the label of this property
     */
    function getLabel() {
        return $this->mLabel;
    }

    /**
     * Sets the label of this property. In wikitext, this is set by the 'Has label' meta property.
     * @param String label the new label for this property
     */
    function setLabel($comment = '') {
        $this->mLabel = $comment;
    }

    /**
     * Returns the label of this property as an annotated wikitext link. It has the form
     * [[Has label::<label string>]] in English.
     * @return String an annotated wikitext link for the label of this property
     */
    function getLabelLink() {
        $tag = $this->getLabelAnnotationTag();
        $label = $this->getLabel();
        return "[[$tag::$label]]";
    }

    /**
     * Returns true if a label has now been defined for this property, where no label was defined in
     * the property's saved article.
     * @return Boolean true if a label has now been added, false otherwise
     */
     function isLabelAdded() {
        return $this->mLabel && !$this->mSavedLabel;
    }

    /**
     * Returns true if no label is defined for this property, where there was a label defined in the
     * property's saved article.
     * @return Boolean true if the label has now been removed, false otherwise
     */
    function isLabelRemoved() {
        return !$this->mLabel && $this->mSavedLabel;
    }

    /**
     * Returns true if a label is defined for this property, where there was a different label
     * defined in the property's saved article.
     * @return Boolean true if the label has been changed, false otherwise
     */
    function isLabelChanged() {
        // no overlap with isLabelAdded or isLabelRemoved
        return $this->mLabel && $this->mSavedLabel && $this->mLabel != $this->mSavedLabel;
    }

    /**
     * Returns the comment of this property. In wikitext, this is set by the 'Has comment' meta
     * property.
     * @return String the comment of this property
     */
    function getComment() {
        return $this->mComment;
    }

    /**
     * Sets the comment of this property. In wikitext, this is set by the 'Has comment' meta
     * property.
     * @param String comment the new comment for this property
     */
    function setComment($comment = '') {
        $this->mComment = $comment;
    }

    /**
     * Returns the comment of this property as an annotated wikitext link. It has the form
     * [[Has comment::<comment string>]] in English.
     * @return String an annotated wikitext link for the comment of this property
     */
    function getCommentLink() {
        $tag = $this->getCommentAnnotationTag();
        $comment = $this->getComment();
        return "[[$tag::$comment]]";
    }

    /**
     * Returns true if a comment has now been defined for this property, where no comment was
     * defined in the property's saved article.
     * @return Boolean true if a comment has now been added, false otherwise
     */
     function isCommentAdded() {
        return $this->mComment && !$this->mSavedComment;
    }

    /**
     * Returns true if no comment is defined for this property, where there was a comment defined in
     * the property's saved article.
     * @return Boolean true if the comment has now been removed, false otherwise
     */
    function isCommentRemoved() {
        return !$this->mComment && $this->mSavedComment;
    }

    /**
     * Returns true if a comment is defined for this property, where there was a different comment
     * defined in the property's saved article.
     * @return Boolean true if the comment has been changed, false otherwise
     */
    function isCommentChanged() {
        // no overlap with isCommentAdded or isCommentRemoved
        return $this->mComment && $this->mSavedComment && $this->mComment != $this->mSavedComment;
    }

    /**
     * Returns the default form (more precisely, the default form's simple name) of this property.
     * In wikitext, this is set by the 'Has default form' meta property. The 'Form' namespace prefix
     * is not included.
     * @return String the default form (name) of this property
     */
    function getDefaultForm() {
        return $this->mDefaultForm;
    }

    /**
     * Sets the default form (more precisely, the default form's simple name) of this property. In
     * wikitext, this is set by the 'Has default form' meta property. The 'Form' namespace prefix is
     * not included.
     * @param String defaultForm the new default form (name) for this property
     */
    function setDefaultForm($label = '') {
        $this->mDefaultForm = $label;
    }

    /**
     * Returns the default form of this property as an annotated wikitext link. It has the form
     * [[Has default form::Form:<form name>]] in English.
     * @return String an annotated wikitext link to the default form of this property
     */
    function getDefaultFormLink() {
        $tag = $this->getDefaultFormAnnotationTag();
        $sfContLang = SFFGlobals::getSFContLang();
        $namespaceLabels = $sfContLang->getNamespaces();
        $formNamespace = $namespaceLabels[SF_NS_FORM];
        $form = $this->getDefaultForm();
        return "[[$tag::$formNamespace:$form]]";
    }

    /**
     * Returns true if a default form has now been defined for this property, where no default form
     * was defined in the property's saved article.
     * @return Boolean true if a default form has now been added, false otherwise
     */
     function isDefaultFormAdded() {
        return $this->mDefaultForm && !$this->mSavedDefaultForm;
    }

    /**
     * Returns true if no default form is defined for this property, where there was a default form
     * defined in the property's saved article.
     * @return Boolean true if the default form has now been removed, false otherwise
     */
    function isDefaultFormRemoved() {
        return !$this->mDefaultForm && $this->mSavedDefaultForm;
    }

    /**
     * Returns true if a default form is defined for this property, where there was a different
     * default form defined in the property's saved article.
     * @return Boolean true if the default form has been changed, false otherwise
     */
    function isDefaultFormChanged() {
        // no overlap with isDefaultFormAdded or isDefaultFormRemoved
        return $this->mDefaultForm && $this->mSavedDefaultForm && $this->mDefaultForm != $this->mSavedDefaultForm;
    }

    /**
     * Returns an array of values allowed by this property. In wikitext, these are set by the
     * 'Allows values' parser function call. The array is not guaranteed to contain only String
     * values.
     * @return array the values allowed by this property.
     */
    function getAllowedValues() {
        return $this->mAllowedValues; // @todo does this work when mAllowedValues is null?
    }

    /**
     * Sets the values allowed by this property. In wikitext, these are set by the 'Allows values'
     * parser function call. Duplicate items (case sensitive) will be removed. The argument's
     * elements are not verified to be of type String.
     * @param array $allowedValues the (possibly null or empty) array of String values to be
     * allowed.
     */
    function setAllowedValues(array $allowedValues = null) {
        $this->mAllowedValues = array_unique($allowedValues);
    }

    /**
     * Returns the allowed values of this property as a wikitext parser function call. It has the
     * form {{#allows values:<value1>|<value2>|<...>}} in English.
     * @return String a wikitext parser function call for the allowed values of this property
     */
    function getAllowsValuesFunctionCall() {
        $tag = $this->getAllowsValuesCallTag();
        $values = $this->getAllowedValues();
        return '{{' . $tag . ':' . implode('|', $values) . '}}'; // clumsy, but cannot escape '{'
    }

    /**
     * Returns true if one or more allowed values have now been defined for this property, where no
     * allowed values were defined in the property's saved article.
     * @return Boolean true if one or more allowed values have now been defined, false otherwise
     */
    function isAllowedValuesAdded() {
        return (count($this->mAllowedValues) > 0 && count($this->mSavedAllowedValues) == 0);
    }

    /**
     * Returns true if no allowed values are defined for this property, where there were one or more
     * allowed values defined in the property's saved article.
     * @return Boolean true if the allowed values have now been removed, false otherwise
     */
    function isAllowedValuesRemoved() {
        return (count($this->mAllowedValues) == 0 && count($this->mSavedAllowedValues) > 0);
    }

    /**
     * Returns true if one or more allowed values are defined for this property, where there was a
     * different set of allowed values defined in the property's saved article.
     * @return Boolean true if the allowed values have been changed, false otherwise
     */
    function isAllowedValuesChanged() {
        // no overlap with isAllowedValuesAdded or isAllowedValuesRemoved
        return (count($this->mAllowedValues) > 0 && count($this->mSavedAllowedValues) > 0 && $this->getAllowedValues() != $this->mSavedAllowedValues);
    }

    /**
     * Returns the title object for this property.
     * @return Title the title object for this property
     */
    function getTitle() {
        return $this->mTitle;
    }

    /**
     * Returns the article object for this property.
     * @return Article the article object for this property
     */
    function getArticle() {
        return new Article($this->mTitle); // @todo what if mTitle is null?
    }

    /**
     * Returns the original (saved) wikitext from this property's article, before any changes made
     * to its meta properties.
     * @return String the original wikitext from this property's article
     */
    function getSavedWikiText() {
        return $this->getArticle()->getContent(); // @todo what if getArticle returns null?
    }

    /**
     * Returns true if an article for this property already exists, false otherwise.
     * @return Boolean true if an article for this property exists, false otherwise
     */
    function hasArticle() {
        // Title::isKnown returns an error if name is empty, so check first
        if ($this->mName == '') {
            return false;
        }
        return $this->mTitle->isKnown();
    }

    /**
     * Returns the (internationalized) link annotation tag used for the type meta property.
     * @return String the annotation tag for the type meta property
     */
    static function getTypeAnnotationTag() {
//        return self::getAnnotationTag('_TYPE', SMW_SP_HAS_TYPE);
        return self::getAnnotationTag('_TYPE');
    }

    /**
     * Returns the (internationalized) link annotation tag used for the Has range meta property.
     * @return String the annotation tag for the Has range meta property
     */
    static function getRangeAnnotationTag() {
        // @todo internationalize?
        return 'Has range';
    }

    /**
     * Returns the (internationalized) parser function tag used for the range parser function call.
     * This is the wikitext name of the parser function for the range.
     * @return String the parser function tag for the Has range meta property
     */
     static function getRangeCallTag() {
        // @todo internationalize?
        return '#range';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the domain meta property.
     * @return String the annotation tag for the range meta property
     */
    static function getDomainAnnotationTag() {
        // @todo internationalize?
        return 'Has domain';
    }

    /**
     * Returns the (internationalized) parser function tag used for the domain parser function call.
     * This is the wikitext name of the parser function for the domain.
     * @return String the parser function tag for the Has domain meta property
     */
     static function getDomainCallTag() {
        // @todo internationalize?
        return '#domain';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the inverse of meta property.
     * @return String the annotation tag for the range meta property
     */
    static function getInverseOfAnnotationTag() {
        // @todo internationalize?
        return 'Is inverse of';
    }

    /**
     * Returns the (internationalized) parser function tag used for the inverse of parser function
     * call. This is the wikitext name of the parser function for the Is inverse of meta property.
     * @return String the parser function tag for the Is inverse of meta property
     */
     static function getInverseOfCallTag() {
        // @todo internationalize?
        return '#inverse of';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the symmetric meta property.
     * @return String the annotation tag for the range meta property
     */
    static function getSymmetricAnnotationTag() {
        // @todo internationalize?
        return 'Is symmetric';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the label meta property.
     * @return String the annotation tag for the range meta property
     */
    static function getLabelAnnotationTag() {
        // @todo internationalize?
        return 'Has label';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the comment meta property.
     * @return String the annotation tag for the range meta property
     */
    static function getCommentAnnotationTag() {
        // @todo internationalize?
        return 'Has comment';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the default form meta property.
     * @return String the annotation tag for the default form meta property
     */
    static function getDefaultFormAnnotationTag() {
        // @todo internationalize?
        return 'Has default form';
    }

    /**
     * Returns the (internationalized) link annotation tag used for the allows value meta property.
     * This is a wikitext link annotation for a single allowed value.
     * @return String the annotation tag for the allows value meta property
     */
    static function getAllowsValueAnnotationTag() {
//        return self::getAnnotationTag('_PVAL', SMW_SP_POSSIBLE_VALUE);
        return self::getAnnotationTag('_PVAL');
    }

    /**
     * Returns the (internationalized) parser function tag used for the allows values parser
     * function call. This is the wikitext name of the parser function for multiple allowed values.
     * @return String the parser function tag for the allowed values meta property
     */
     static function getAllowsValuesCallTag() {
        // @todo internationalize?
        return '#allows values';
    }

    /**
     * Returns the internationalized link annotation tag for a given special meta property.
     * @param String $propertyLabel the property label as defined starting from SMW 1.4
     * @param Integer $specialProperty the special property constant as used in earlier SMW versions
     * @return String the link annotation tag
     */
//    private static function getAnnotationTag($propertyLabel, $specialProperty) {
//        $smwContLang = SFFGlobals::getSMWContLang();
//        // handling of special property labels changed in SMW 1.4
//        if (method_exists($smwContLang, 'getPropertyLabels')) {
//            $propertyLabels = $smwContLang->getPropertyLabels();
//            return $propertyLabels[$propertyLabel];
//        } else {
//            $specialProperties = $smwContLang->getSpecialPropertiesArray();
//            return $specialProperties[$specialProperty];
//        }
//
//    }
    private static function getAnnotationTag($propertyLabel) {
        $smwContLang = SFFGlobals::getSMWContLang();
        // handling of special property labels changed in SMW 1.4
        $propertyLabels = $smwContLang->getPropertyLabels();
        return $propertyLabels[$propertyLabel];
    }

}
?>
