<?php

if (!defined('MEDIAWIKI')) die();

/**
 * The SFFProperty class represents a SMW Property.
 *
 * @author Henri Gerrits
 */
class SFFTemplate {

    // instance variables - see doc on getters/setters
    // mSavedXxx are retrieved from the existing article
    // mXxx are the same, except where overridden by the request parameters, i.e., the new values
    private $mName;
    private $mCategory;
    private $mFields;
    private $mAggregatingProperty;
    private $mAggregationLabel;
    private $mTemplateFormat;

    // transient variables
    private $mTitle;


    /**
     * Constructs a new SFFTemplate. The new values for the properties are taken from the
     * request parameters, if present. The saved values of the properties are loaded from the
     * article.
     * @param WebRequest $request optional web request containing new values for the properties
     */
    function __construct(WebRequest $request = null) {
        if (isset($request)) {
            $this->setName($request->getVal('name'));
            $this->setTemplateFormat('standard');
            if ($this->hasArticle()) {
                // get the original values from the wikitext
                $this->loadSavedProperties();
            }
            // set 'new' values from the request parameters
            if ($request->getCheck('category')) {
                $this->setCategory($request->getVal('category'));
            }

            // cycle through the query values, setting the appropriate local variables
            $fields = array();
            $updateFields = false;
            foreach ($request->getValues() as $parameter => $label) {
                $parameterNameComponents = explode("_", $parameter);
                // we only care about query parameters of the form "a_b"
                if (count($parameterNameComponents) != 2)
                    continue;
                list ($parameterName, $index) = $parameterNameComponents;
                if ($parameterName == 'name') {
                    $updateFields = true;
                    $semanticProperty = $request->getVal('semanticproperty_' . $index);
                    if ($index != 'new' || ($index == 'new' && ($label || $semanticProperty))) {
                        if ($request->getVal('del_' . $index) != '') {
                            // do nothing - this field won't get added to the new list
                        } else {
                            $field = SFFTemplateField::create($label, $request->getVal('label_' . $index));
                            $field->semantic_property = $request->getVal('semanticproperty_' . $index);
                            $field->is_list = $request->getCheck('is_list_' . $index);
                            $fields[] = $field;
                        }
                    }
                }
            }
            // now check whether to load the properties for the category
            if ($request->getVal('loadproperties')) {
                $propertieNames = $this->getPropertiesForCategory();
                foreach ($propertieNames as $propertyName) {
                    $label = SFFUtils::getLabelForProperty($propertyName);
                    $field = SFFTemplateField::create($label, $label);
                    $field->semantic_property = $propertyName;
                    $field->is_list = false;
                    $fields[] = $field;
                    $updateFields = true;
                }
            }
            if ($updateFields) {
                $this->setFields($fields);
            }
            if ($request->getCheck('semanticproperty_aggregation')) {
                $this->setAggregatingProperty($request->getVal('semanticproperty_aggregation'));
            }
            if ($request->getCheck('aggregationlabel')) {
                $this->setAggregationLabel($request->getVal('aggregationlabel'));
            }
            if ($request->getCheck('templateformat')) {
                $this->setTemplateFormat($request->getVal('templateformat'));
            }
        }
    }

    private function getPropertiesForCategory() {
        $categoryName = $this->getCategory();
        if ($categoryName) {
            return SFFUtils::getPropertiesForCategory($categoryName);
        } else {
            return array();
        }
    }
    /**
     * Reads the saved property values.
     */
    private function loadSavedProperties() {
        // this assumes that $mName has already been set

        $article = new Article($this->getTitle());
        $templateText = $article->getContent();
        // ignore 'noinclude' sections and 'includeonly' tags
        $templateText = StringUtils::delimiterReplace('<noinclude>', '</noinclude>', '', $templateText);
        $templateText = strtr($templateText, array('<includeonly>' => '', '</includeonly>' => ''));

        $this->loadCategoryField($templateText);
        $this->loadSemanticFields($templateText);
        $this->loadAggregationFields($templateText);
        $this->loadTemplateFormat($templateText);
    }

    /**
     * Reads the category field from the template text.
     * @param String $templateText the 'includeonly' part of the template text
     */
    private function loadCategoryField($templateText) {

        // Example:
        // [[Category::Some category]]
        $regex = '/\[\['; // "[["
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= 'Category'; // "Category" @todo: internationalize
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= ':'; // ":"
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '('; // start group 1 (category name)
        $regex .= '.+?'; // one or more characters, not greedy
        $regex .= ')'; // end group 1
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '\]\]'; // "]]"
        $regex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        if (preg_match_all($regex, $templateText, $matches)) {
            $this->setCategory(end($matches[1]));
        }

    }


    /**
     * Reads the semantic fields from the template text.
     * @param String $templateText the 'includeonly' part of the template text
     */
    private function loadSemanticFields($templateText) {
        $templateFields = array();
        $fieldNamesArray = array();

        // Get the fields of the template, both semantic and otherwise, by parsing
        // the text of the template.
        // The way this works is that fields are found and then stored in an
        // array based on their location in the template text, so that they
        // can be returned in the order in which they appear in the template, even
        // though they were found in a different order.
        // Some fields can be found more than once (especially if they're part
        // of an "#if" statement, so they're only recorded the first time they're
        // found. Also, every field gets replaced with a string of x's after
        // being found, so it doesn't interfere with future parsing.

        $sffRegex = '/{{\s*#sffheadrow\s*:\s*([^!|]*)\s*}}\s*{{#arraymap:{{{([^|}]*:?[^|}]*)[^\[]*\[\[([^:=]*:?[^:=]*)(:[:=])/mis';
        $sfRegex = '/!\s*([^!|]*)\s*\|\s*{{#arraymap:{{{([^|}]*:?[^|}]*)[^\[]*\[\[([^:=]*:?[^:=]*)(:[:=])/mis';

        // first, look for SFF style "arraymap" parser function calls that map a
        // relation or attribute onto a list
        $positiveMatch = preg_match_all($sffRegex, $templateText, $matches);
        // if no positive match, test the SF style "arraymap" parser function calls
        if (!$positiveMatch) {
            $positiveMatch = preg_match_all($sfRegex, $templateText, $matches);
        }
        if ($positiveMatch) {
            // this is a two-dimensional array; we need the last three of the four
            // sub-arrays; we also have to remove redundant values
            foreach ($matches[2] as $i => $fieldName) {
                $semanticProperty = $matches[3][$i];
                $fullFieldText = $matches[0][$i];
                if (!in_array($fieldName, $fieldNamesArray)) {
                    $fieldLabel = $matches[1][$i];
                    $templateField = SFFTemplateField::create($fieldName, ucfirst($fieldLabel));
                    $templateField->setSemanticProperty($semanticProperty);
                    $templateField->is_list = true;
                    $curPos = stripos($templateText, $fullFieldText);
                    $templateFields[$curPos] = $templateField;
                    $fieldNamesArray[] = $fieldName;
                    $replacement = str_repeat("x", strlen($fullFieldText));
                    $templateText = str_replace($fullFieldText, $replacement, $templateText);
                }
            }
        }

        // second, look for normal relation/attribute calls
        $sfRegex = '/!'; // "!"
        $sfRegex .= '\s*'; // some whitespace (greedy)
        $sfRegex .= '('; // start group 1 (label)
        $sfRegex .= '[^!|]*'; // zero or more characters, except "!" and "|"
        $sfRegex .= ')'; // end group 1
        $sfRegex .= '\s*'; // some whitespace (greedy)
        $sfRegex .= '\|'; // "|"
        $sfRegex .= '\s*'; // some whitespace (greedy)
        $sfRegex .= '\[\['; // "[["
        $sfRegex .= '('; // start group 2 (semantic property name)
        $sfRegex .= '[^:=]*'; // zero or more characters, except ":" and "="
        $sfRegex .= ':*?'; // zero or more ":", not greedy
        $sfRegex .= ')'; // end group 2
        $sfRegex .= '('; // start group 3 (separator)
        $sfRegex .= ':[:=]'; // "::" or ":="
        $sfRegex .= ')'; // end group 3
        $sfRegex .= '{{{'; // "{{{"
        $sfRegex .= '('; // start group 4 (field name reference)
        $sfRegex .= '[^\]\|}]*'; // zero or more characters, except "]", "|" and "}"
        $sfRegex .= ')'; // end group 4
        $sfRegex .= '.*?'; // zero or more characters, not greedy (i.e., anything after and including starting "]", "|" or "}"
        $sfRegex .= '\]\]'; // "]]"
        $sfRegex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        $sffRegex = '/{{'; // "{!}{"
        $sffRegex .= '\s*'; // some whitespace (greedy)
        $sffRegex .= '#sffheadrow'; // "#SFFHeadRow"
        $sffRegex .= '\s*'; // some whitespace (greedy)
        $sffRegex .= ':'; // ":"
        $sffRegex .= '\s*'; // some whitespace (greedy)
        $sffRegex .= '('; // start group 1 (label)
        $sffRegex .= '[^!|]*'; // zero or more characters, except "!" and "|"
        $sffRegex .= ')'; // end group 1
        $sffRegex .= '\s*'; // some whitespace (greedy)
        $sffRegex .= '\}}'; // "}}"
        $sffRegex .= '\s*'; // some whitespace (greedy)
        $sffRegex .= '\[\['; // "[["
        $sffRegex .= '('; // start group 2 (semantic property name)
        $sffRegex .= '[^:=]*'; // zero or more characters, except ":" and "="
        $sffRegex .= ':*?'; // zero or more ":", not greedy
        $sffRegex .= ')'; // end group 2
        $sffRegex .= '('; // start group 3 (separator)
        $sffRegex .= ':[:=]'; // "::" or ":="
        $sffRegex .= ')'; // end group 3
        $sffRegex .= '{{{'; // "{{{"
        $sffRegex .= '('; // start group 4 (field name reference)
        $sffRegex .= '[^\]\|}]*'; // zero or more characters, except "]", "|" and "}"
        $sffRegex .= ')'; // end group 4
        $sffRegex .= '.*?'; // zero or more characters, not greedy (i.e., anything after and including starting "]", "|" or "}"
        $sffRegex .= '\]\]'; // "]]"
        $sffRegex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        // test for SFF style semantic fields
        $positiveMatch = preg_match_all($sffRegex, $templateText, $matches);
        // if no positive match, test the SF style semantic fields
        if (!$positiveMatch) {
            $positiveMatch = preg_match_all($sfRegex, $templateText, $matches);
        }
        if ($positiveMatch) {
            // this is a two-dimensional array; we need the last three of the four
            // sub-arrays; we also have to remove redundant values
            foreach ($matches[2] as $i => $semanticProperty) {
                $fullFieldText = $matches[0][$i];
                $fieldName = $matches[4][$i];
                if (!in_array($fieldName, $fieldNamesArray)) {
                    $fieldLabel = $matches[1][$i];
                    $templateField = SFFTemplateField::create($fieldName, ucfirst($fieldLabel));
                    $templateField->setSemanticProperty($semanticProperty);
                    $curPos = stripos($templateText, $fullFieldText);
                    $templateFields[$curPos] = $templateField;
                    $fieldNamesArray[] = $fieldName;
                    $replacement = str_repeat("x", strlen($fullFieldText));
                    $templateText = str_replace($fullFieldText, $replacement, $templateText);
                }
            }
        }

        // finally, get any non-semantic fields defined
        if (preg_match_all('/{{{([^|}]*)/mis', $templateText, $matches)) {
            foreach ($matches[1] as $i => $fieldName) {
                $fullFieldText = $matches[0][$i];
                if (($fullFieldText != '') && ($fieldName != 'SFForm') && (!in_array($fieldName, $fieldNamesArray))) {
                    $curPos = stripos($templateText, $fullFieldText);
                    $templateFields[$curPos] = SFFTemplateField::create($fieldName, ucfirst($fieldName));
                    $fieldNamesArray[] = $fieldName;
                }
            }
        }
        ksort($templateFields);
        $this->setFields($templateFields);
    }

    /**
     * Reads the aggregation fields from the template text.
     * @param String $templateText the 'includeonly' part of the template text
     */
    private function loadAggregationFields($templateText) {
        // Example:
        // ! Title for list
        // | {{#ask:[[PropertyName::{{SUBJECTPAGENAME}}]]|format=list}}
        $regex = '/!'; // "!"
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '('; // start group 1 (aggregation label)
        $regex .= '[^!|]+'; // one or more characters, except "!" and "|"
        $regex .= ')'; // end group 1
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '\|'; // "|"
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '{{#ask:\[\['; // "{{#ask:[["
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '('; // start group 2 (aggregating property name)
        $regex .= '.+?'; // one or more characters, not greedy
        $regex .= ')'; // end group 2
        $regex .= '\s*'; // some whitespace (greedy)
        $regex .= '::{{SUBJECTPAGENAME}}\]\]\|format=list}}'; // "::{{SUBJECTPAGENAME}}]]|format=list}}"
        $regex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        if (preg_match($regex, $templateText, $matches)) {
            $this->setAggregationLabel($matches[1]);
            $this->setAggregatingProperty($matches[2]);
        }
    }

    /**
     * Reads the template format field from the template text.
     * @param String $templateText the 'includeonly' part of the template text
     */
    private function loadTemplateFormat($templateText) {
        // Example for standard format:
        // {| class="wikitable"
        $standardRegex = '/\{\|'; // "{|"
        $standardRegex .= '\s*'; // some whitespace (greedy)
        $standardRegex .= 'class'; // "class"
        $standardRegex .= '\s*'; // some whitespace (greedy)
        $standardRegex .= '='; // "="
        $standardRegex .= '\s*'; // some whitespace (greedy)
        $standardRegex .= '"wikitable"'; // ""wikitable""
        $standardRegex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        // Example for infobox format:
        // {| style="width: 30em; font-size: 90%; border: 1px solid #aaaaaa;
        // background-color: #f9f9f9; color: black; margin-bottom: 0.5em; margin-left: 1em;
        // padding: 0.2em; float: right; clear: right; text-align:left;"
        // ! style="text-align: center; background-color:#ccccff;" colspan="2" |<big>{{PAGENAME}}</big>
        //
        // the "float: right" is the key style that characterizes the infobox format

        $infoboxRegex = '/\{\|'; // "{|"
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= 'style'; // "style"
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= '='; // "="
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= '"'; // """
        $infoboxRegex .= '[^"]*'; // some text except """
        $infoboxRegex .= 'float'; // "float"
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= ':'; // ":"
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= 'right'; // "right"
        $infoboxRegex .= '\s*'; // some whitespace (greedy)
        $infoboxRegex .= ';'; // ";"
        $infoboxRegex .= '[^"]*'; // some text except """
        $infoboxRegex .= '"'; // """
        $infoboxRegex .= '/mis'; // multiline, case independent, dotall (. includes newlines)

        if (preg_match($infoboxRegex, $templateText, $matches)) {
            $this->setTemplateFormat('infobox');
        }
    }

    /**
     * Returns the name of this template.
     * @return String the name of this template
     */
    function getName() {
        return $this->mName;
    }

    /**
     * Sets the name of this template. Excess whitespace is removed, and underscores are replaced by
     * spaces.
     * @param String $name the name of this template
     */
    function setName($name = '') {
        // reduce any whitespace or underscores between words to a single space
        $name = preg_replace('/(\s|_)+/', ' ', $name);
        $this->mName = trim($name);
        // cache title object
        $this->mTitle = Title::makeTitleSafe(NS_TEMPLATE, $this->getName());
    }

    /**
     * Returns the category (more precisely, the category's name) defined by this template.
     * @return String the category name defined by this template
     */
    function getCategory() {
        return $this->mCategory;
    }

    /**
     * Sets the category (more precisely, the category's name) defined by this template.
     * @param String $category the category defined by this template
     */
    function setCategory($category = '') {
        $this->mCategory = $category;
    }

    /**
     * Returns the semantic fields used by this template.
     * @return array(SFFTemplateField) the template fields used by this template
     */
    function getFields() {
        return $this->mFields;
    }

    /**
     * Sets the semantic fields used by this template.
     * @param <rray(SFFTemplateField) $fields the template fields used by this template
     */
    function setFields($fields) {
        $this->mFields = $fields;
    }

    /**
     * Returns the aggregating property (more precisely, the property's name). Used to list, on any
     * page using this template, all of the articles that have this aggregating property pointing to
     * that page.
     * @return String the aggregating property name
     */
    function getAggregatingProperty() {
        return $this->mAggregatingProperty;
    }

    /**
     * Sets the aggregating property (more precisely, the property's name). Used to list, on any
     * page using this template, all of the articles that have this aggregating property pointing to
     * that page.
     * @param String $property the aggregating property name
     */
    function setAggregatingProperty($property = '') {
        $this->mAggregatingProperty = $property;
    }

    /**
     * Returns the aggregation label. The aggregation label is the title for the list of articles
     * pointing, using the aggregation property, to any article that uses this template.
     * @return String the aggregation label
     */
    function getAggregationLabel() {
        return $this->mAggregationLabel;
    }

    /**
     * Sets the aggregation label. The aggregation label is the title for the list of articles
     * pointing, using the aggregation property, to any article that uses this template.
     * @param String $label the aggregation label
     */
    function setAggregationLabel($label = '') {
        $this->mAggregationLabel = $label;
    }

    /**
     * Returns the template format used by articles to display the semantic properties.
     * @return String the template format
     */
    function getTemplateFormat() {
        return $this->mTemplateFormat;
    }

    /**
     * Sets the template format used by articles to display the semantic properties.
     * @param String $format the template format
     */
    function setTemplateFormat($format = '') {
        $this->mTemplateFormat = $format;
    }

    /**
     * Returns the title object for this template.
     * @return Title the title object for this template
     */
    function getTitle() {
        return $this->mTitle;
    }

    /**
     * Returns true if an article for this template already exists, false otherwise.
     * @return Boolean true if an article for this template exists, false otherwise
     */
    function hasArticle() {
        // Title::isKnown returns an error if name is empty, so check first
        if ($this->mName == '') {
            return false;
        }
        return $this->mTitle->isKnown();
    }

 }
?>