<?php
/**
 * Description of SFF_PropertyParserFunctions
 * Following the example of http://www.mediawiki.org/wiki/Manual:Parser_functions
 *
 * @author Rens te Kloese
 */
class SFFFormParserFunctions {

    /**
     * Sets up the parser function. Sets a function hook for the 'allows values' magic word.
     */
    static function formParserFunctions_Setup() {
        $parser = SFFGlobals::getParser();
        // Set a function hook associating the 'allows values' magic word with our function
        $parser->setFunctionHook('sff_forminput', 'SFFFormParserFunctions::renderSFFFormInput');
    }

    /**
     * Registers the 'allows values' magic word with the MediaWiki server.
     * @param array $magicWords the array of magic words to which it will be added
     * @param String $langCode the language code
     * @return Boolean always returns true
     */
    static function formParserFunctions_Magic(&$magicWords, $langCode) {
        // Add the magic word
        // The first array element is case sensitive, in this case it is not case sensitive
        // All remaining elements are synonyms for our parser function
        $magicWords['sff_forminput'] = array(0, 'sff_forminput');
        // unless we return true, other parser functions extensions won't get loaded.
        return true;
    }

    /**
     * The parser function itself. Expands the 'allows values' function call to an unordered list
     * (HTML) of annotated 'allows value' wikitext links.
     * @param array anonymous array containing the parser object at [0] and the arguments to the
     * function call in subsequent positions (wikitext with templates expanded)
     * @return String a unordered list (HTML) of annotated 'allows value' wikitext links
     *
     * Copy from the source: SF_ParserFunctions.php
     */
    static function renderSFFFormInput (&$parser, $inFormName = '', $inSize = '25', $inValue = '', $inButtonStr = '', $inQueryStr = '') {
        $ap = SpecialPage::getPage('SFF_AddPage');
        $ap_url = $ap->getTitle()->getLocalURL();
        $str = <<<END
            <form action="$ap_url" method="get">
  <p><input type="text" name="page_name" size="$inSize" value="$inValue">

END;
        // if the add page URL looks like "index.php?title=Special:AddPage"
        // (i.e., it's in the default URL style), add in the title as a
        // hidden value
        if (($pos = strpos($ap_url, "title=")) > -1) {
            $str .= '			<input type="hidden" name="title" value="' . urldecode(substr($ap_url, $pos + 6)) . '">' . "\n";
        }
        if ($inFormName == '') {
            $str .= SFUtils::formDropdownHTML();
        } else {
            $str .= '			<input type="hidden" name="form" value="' . $inFormName . '">' . "\n";
        }
        // recreate the passed-in query string as a set of hidden variables
        $query_components = explode('&', $inQueryStr);
        foreach ($query_components as $component) {
            $subcomponents = explode('=', $component, 2);
            $key = (isset($subcomponents[0])) ? $subcomponents[0] : '';
            $val = (isset($subcomponents[1])) ? $subcomponents[1] : '';
            $str .= '			<input type="hidden" name="' . $key . '" value="' . $val . '">' . "\n";
        }
        wfLoadExtensionMessages('SemanticForms');
        $button_str = ($inButtonStr != '') ? $inButtonStr : wfMsg('addoreditdata');
        $str .= <<<END
            <input type="submit" value="$button_str"></p>
            </form>
END;
        return array($str, 'noparse' => true, 'isHTML' => true);
    }

    /**
     * Function is copied from the method "onParserAfterTidy" from the class SMW_ParseData.php
     *
     * Hook function fetches category information and other final settings from parser output,
     * so that they are also replicated in SMW for more efficient querying.
     */
    static public function SFFonParserAfterTidy(&$parser, &$text) {
        // begin
        global $wgOut, $wgUser;

        // init variables
        $sk = $wgUser->getSkin();
        $properties = array();
        $propertiesInverseOf = array();
        $propertiesIsSymmetric = array();
        $title = $parser->mTitle;
        $ns = $parser->mTitle->getNamespace();
        if ($ns == 0){
            $article = new Article($title);
            // determine article templates
            $templates = $article->getUsedTemplates();
            foreach ($templates as $pgTemplate) {
                $article = new Article($pgTemplate);
                $templateTitle = $article->getTitle();
                $tif = SFFTemplateInForm::create($templateTitle->getText(), '', false);
                $fields = $tif->getAllFields();
                foreach ($fields as $field) {
                    $properties[] = $field->semantic_property;
                }
            }
            // check each property if isSemmetric or inverseOf
            foreach ($properties as $artProperty) {
                $newRequest = new WebRequest();
                $newRequest->setVal('name',$artProperty);
                $webproperty = new SFFProperty($newRequest);
                $type = $webproperty->getType();
                if ($type == 'Page') {
                    //Only type Page allowed
                    $inverseOf = $webproperty->getInverseOf();
                    $isSymmetric = $webproperty->getSymmetric();
                    $propertyDomain = $webproperty->getDomain();
                    $propertyRange = $webproperty->getRange();
                    //Fill table with property field values
                    $ownValues = &smwfGetStore()->getPropertyValues($title, SMWPropertyValue::makeUserProperty($artProperty));

                    if ( count($ownValues) == 0 ) {
                        $ownValues[] = SMWPropertyValue::makeUserProperty($title->getText() );
                    }
                    //If inverseOf property
                    if ( count($inverseOf) > 0 ) {
                        foreach ($inverseOf as $invProp) {
                            $found = 0;
                            foreach ($propertiesInverseOf as $prop) {
                                if ($invProp == $prop) $found = 1;
                            }
                            if ($found == 0) {
                                $propertiesInverseOf[] = $invProp;
                                //Determine implied values for property:
                                $invPages = array();
                                $inverseOfPages = SFFArticleFunctions::listPagesWithPropertyValues($title->getText(),
                                    SMWPropertyValue::makeUserProperty($invProp));
                                $count = count($inverseOfPages);
                                if ( count($inverseOfPages) > 0 ) {
                                    foreach($inverseOfPages as $singlepage) {
                                        if ($singlepage->getLongHTMLText() != $title) {
                                            //skip own page
                                            $found = 0;
                                            foreach ($invPages as $invPage) {
                                                if ($singlepage->getLongHTMLText() == $invPage) $found = 1;
                                            }
                                            foreach ($ownValues as $ownPage) {
                                                if ($singlepage->getLongHTMLText() == $ownPage->getLongHTMLText()) $found = 1;
                                            }
                                            if ($found == 0) {$invPages[] = $singlepage->getLongHTMLText();}
                                        }
                                    }
                                }
                                if (count($invPages) > 0) {
                                    $wgOut->addHTML('<b>' . $artProperty . ' (implied) :</b>');
                                    foreach ($invPages as $invPage){
                                        $pageLink = $sk->makeKnownLinkObj( $invPage, $invPage );
                                        $wgOut->addHTML($pageLink . ', ');
                                    }
                                    $wgOut->addHTML('<p>');
                                }
                            }
                        }
                    }
                    //IF symmetric property
                    if ( $isSymmetric == 'true' ) {
                        $symmPages = array();
                        foreach ($ownValues as $page) {//values of property for this page
                            $title2 = $page->getShortWikiText();
                            $isSymmetricPages = SFFArticleFunctions::listPagesWithPropertyValues($title2,
                                SMWPropertyValue::makeUserProperty($artProperty));
                            foreach($isSymmetricPages as $singlepage) {
                                if ($singlepage->getLongHTMLText() != $title ) {
                                    //skip own page
                                    $found = 0;
                                    foreach ($symmPages as $symmPage) {
                                        if ($singlepage->getLongHTMLText() == $symmPage) $found = 1;
                                    }
                                    foreach ($ownValues as $ownPage) {
                                        if ($singlepage->getLongHTMLText() == $ownPage->getLongHTMLText()) $found = 1;
                                    }
                                    if ($found == 0) {$symmPages[] = $singlepage->getLongHTMLText();}
                                }
                            }
                        }
                        if (count($symmPages) > 0) {
                            $wgOut->addHTML('<b>' . $artProperty . ' (implied) :</b> ');
                            foreach ($symmPages as $symmPage){
                                //$wgOut->addHTML($symmPage . ', ');
                                $pageLink = $sk->makeKnownLinkObj( $invPage, $invPage );
                                $wgOut->addHTML($pageLink . ', ');
                            }
                            $wgOut->addHTML('<p>');
                        }
                    }
                }
            }
        }
        // end
        if (SMWParseData::getSMWData($parser) === NULL) return true;
        $categories = $parser->mOutput->getCategoryLinks();
        foreach ($categories as $name) {
            $pinst = SMWPropertyValue::makeProperty('_INST');
            $dv = SMWDataValueFactory::newPropertyObjectValue($pinst);
            $dv->setValues($name,NS_CATEGORY);
            SMWParseData::getSMWData($parser)->addPropertyObjectValue($pinst,$dv);
            if (SMWParseData::getSMWData($parser)->getSubject()->getNamespace() == NS_CATEGORY) {
                $psubc = SMWPropertyValue::makeProperty('_SUBC');
                $dv = SMWDataValueFactory::newPropertyObjectValue($psubc);
                $dv->setValues($name,NS_CATEGORY);
                SMWParseData::getSMWData($parser)->addPropertyObjectValue($psubc,$dv);
            }
        }
        $sortkey = ($parser->mDefaultSort?$parser->mDefaultSort:SMWParseData::getSMWData($parser)->getSubject()->getText());
        SMWParseData::getSMWData($parser)->getSubject()->setSortkey($sortkey);
        return true;
    }


}
?>
