<?php
/**
 * Description of SFF_PropertyParserFunctions
 * Following the example of http://www.mediawiki.org/wiki/Manual:Parser_functions
 *
 * @author Henri
 */
class SFFPropertyParserFunctions {

    /**
     * Sets up the parser function. Sets a function hook for the 'allows values' magic word.
     */
    static function propertyParserFunctions_Setup() {
        $parser = SFFGlobals::getParser();
        // Set a function hook associating the 'allows values' magic word with our function
        $parser->setFunctionHook('sff_allowsvalues', 'SFFPropertyParserFunctions::allowsValuesFunction_Render');
        $parser->setFunctionHook('sff_range', 'SFFPropertyParserFunctions::rangeFunction_Render');
        $parser->setFunctionHook('sff_domain', 'SFFPropertyParserFunctions::domainFunction_Render');
        $parser->setFunctionHook('sff_inverseof', 'SFFPropertyParserFunctions::inverseOfFunction_Render');
    }

    /**
     * Registers the magic words with the MediaWiki server.
     * @param array $magicWords the array of magic words to which it will be added
     * @param String $langCode the language code
     * @return Boolean always returns true
     */
    static function propertyParserFunctions_Magic(&$magicWords, $langCode) {
        // Add the magic words
        // The first array element is case sensitive, in this case it is not case sensitive
        // All remaining elements are synonyms for our parser function
        $magicWords['sff_allowsvalues'] = array(0, 'allows values');
        $magicWords['sff_range'] = array(0, 'range');
        $magicWords['sff_domain'] = array(0, 'domain');
        $magicWords['sff_inverseof'] = array(0, 'inverse of');
        // unless we return true, other parser functions extensions won't get loaded.
        return true;
    }

    /**
     * The allows values parser function itself. Expands the 'allows values' function call to an
     * unordered list (HTML) of annotated 'allows value' wikitext links.
     * @param array anonymous array containing the parser object at [0] and the arguments to the
     * function call in subsequent positions (wikitext with templates expanded)
     * @return String an unordered list (HTML) of annotated 'allows value' wikitext links
     */
    static function allowsValuesFunction_Render() {
        // start at 1 because the parser object is at position 0
        for ($i = 1; $i < func_num_args(); $i++) {
            $allowsvalues[] = func_get_arg($i);
        }
        // @todo simply use comma separation instead (or better, use a switch to define rendering)
        $tag = SFFProperty::getAllowsValueAnnotationTag();
        $output = "<ul>\n";
        foreach ($allowsvalues as $value) {
            $output .= "    <li>[[$tag::$value]]</li>\n";
        }
        $output .= "</ul>\n";
        return $output;
    }

    /**
     * The range parser function itself. Expands the 'range' function call to a comma separated list
     * of annotated 'has range' wikitext links.
     * @param array anonymous array containing the parser object at [0] and the arguments to the
     * function call in subsequent positions (wikitext with templates expanded)
     * @return String a comma separated list of annotated 'has range' wikitext links
     */
    static function rangeFunction_Render() {
        // start at 1 because the parser object is at position 0
        for ($i = 1; $i < func_num_args(); $i++) {
            $range[] = func_get_arg($i);
        }
        $tag = SFFProperty::getRangeAnnotationTag();
        return self::expandToListOfLinks($range, $tag, 'Category');
    }

    /**
     * The domain parser function itself. Expands the 'domain' function call to a comma separated
     * list of annotated 'has domain' wikitext links.
     * @param array anonymous array containing the parser object at [0] and the arguments to the
     * function call in subsequent positions (wikitext with templates expanded)
     * @return String a comma separated list of annotated 'has domain' wikitext links
     */
    static function domainFunction_Render() {
        // start at 1 because the parser object is at position 0
        for ($i = 1; $i < func_num_args(); $i++) {
            $domain[] = func_get_arg($i);
        }
        $tag = SFFProperty::getDomainAnnotationTag();
        return self::expandToListOfLinks($domain, $tag, 'Category');
    }

    /**
     * The inverse of parser function itself. Expands the 'inverse of' function call to a comma
     * separated list of annotated 'is inverse of' wikitext links.
     * @param array anonymous array containing the parser object at [0] and the arguments to the
     * function call in subsequent positions (wikitext with templates expanded)
     * @return String a comma separated list of annotated 'is inverse of' wikitext links
     */
    static function inverseOfFunction_Render() {
        // start at 1 because the parser object is at position 0
        for ($i = 1; $i < func_num_args(); $i++) {
            $inverseOf[] = func_get_arg($i);
        }
        $tag = SFFProperty::getInverseOfAnnotationTag();
        return self::expandToListOfLinks($inverseOf, $tag, 'Property');
    }

    /**
     * Expands an array of names to a comma-separated list of semantic links.
     * @param array $array the array of names
     * @param String $tag the semantic property name to be used
     * @param String $namespace the namespace to prefix the names with
     * @return String a comma-separated list of semantic links
     */
    private static function expandToListOfLinks($array, $tag, $namespace) {
        $links = '';
        foreach ($array as $item) {
            if ($links != '') $links .= ', ';
            $links .= "[[$tag::$namespace:$item|$item]]";
        }
        return $links;
    }

}
?>
