<?php

if (!defined('MEDIAWIKI')) die();


class SFFEditProperty extends SFFPropertySpecialPage {

    /**
     * Constructs the special page and registers it with MW.
     */
    function __construct() {
        SpecialPage::SpecialPage('SFF_EditProperty');
    }

}

class SFFCreateProperty extends SFFPropertySpecialPage {

    /**
     * Constructs the special page and registers it with MW.
     */
    function __construct() {
        SpecialPage::SpecialPage('SFF_CreateProperty');
    }

}


/**
 * A special page providing a form that allows the user to create or edit a semantic property.
 *
 * @author Henri Gerrits
 */
abstract class SFFPropertySpecialPage extends SpecialPage {

    /**
     * Class entry point that displays the appropriate page depending on the circumstances.
     * Overrides SpecialPage::execute().
     * @param <type> $query not used
     */
    function execute($query) {
        wfLoadExtensionMessages('SFF_extension');
        $this->setHeaders();
        $request = SFFGlobals::getRequest();
        $property = new SFFProperty($request);
        if (!$request->getCheck('name')) {
            $this->doChoose();
        } else {
            $isSavePage = $request->getCheck('wpSave');
            $isPreviewPage = $request->getCheck('wpPreview');
            if ($isSavePage || $isPreviewPage) {
                $this->doView($property, $isSavePage, $isPreviewPage);
            } else {
                $this->doEdit($property);
            }
        }
    }

    /**
     * Builds a 'choose' page, containing an option to create a new property, as well as a list of
     * existing properties that can be edited.
     */
    private function doChoose() {
        $outputPage = SFFGlobals::getOutputPage();
        $outputPage->setPageTitle($this->buildPageTitle());

        list($limit, $offset) = wfCheckLimits();
        $queryPage = new SFFPropertyQueryPage();
        $queryPage->doQuery($offset, $limit);
        SMWOutputs::commitToOutputPage($outputPage); // make sure locally collected output data is pushed to the output!
    }


    /**
     * Builds the 'view' page after the save or preview buttons have been pressed.
     * @param SFFProperty $property the property to be viewed on the page
     * @param Boolean $isSavePage true if save button has been pressed previously
     * @param Boolean $isPreviewPage true if preview button has been pressed previously
     */
    private function doView(SFFProperty $property, $isSavePage, $isPreviewPage) {
        if ($property->getName() == '') {
            $this->doEdit($property, wfMsg('sff-editproperty-edit-error-nameblank'));
            return;
        }
        // @todo test whether name for new property already exists
        $outputPage = SFFGlobals::getOutputPage();
        $outputPage->setPageTitle($this->buildPageTitle($property));
        $outputPage->setArticleBodyOnly(true);
        $title = $property->getTitle();
        if ($property->hasArticle()) {
            $fullText = $this->updateExistingPropertyText($property);
        } else {
            $fullText = $this->generateNewPropertyText($property);
        }
        // HTML-encode
        $fullText = str_replace('"', '&quot;', $fullText);
        $text = SFUtils::printRedirectForm($title, $fullText, "", $isSavePage, $isPreviewPage, false, false, false, null, null);
        $outputPage->addHTML($text);
    }


    /**
     * Builds the edit property with form page and places it in the global OutputPage object.
     * @param SFFProperty $property the property to be edited in the form
     * @param String $nameError any errors to be displayed related to the name field
     */
    private function doEdit(SFFProperty $property, $nameError = '') {
        $outputPage = SFFGlobals::getOutputPage();
        $mwContLang = SFFGlobals::getMWContLang();
        $smwContLang = SFFGlobals::getSMWContLang();

        $outputPage->setPageTitle($this->buildPageTitle($property));

        // HG: what is this for????
        $javascript_text =<<<END
function toggleAllowedValues() {
    var values_div = document.getElementById("allowed_values");
}

END;

        $namespaceLabels = $mwContLang->getNamespaces();
        $specialNamespace = $namespaceLabels[NS_SPECIAL];
        $nameLabel = wfMsg('sff-editproperty-edit-name-label');
        $typeLabel = wfMsg('sff-editproperty-edit-type-label');
        $rangeLabel = wfMsg('sff-editproperty-edit-range-label');
        $domainLabel = wfMsg('sff-editproperty-edit-domain-label');
        $inverseOfLabel = wfMsg('sff-editproperty-edit-inverseof-label');
        $symmetricLabel = wfMsg('sff-editproperty-edit-symmetric-label');
        $labelLabel = wfMsg('sff-editproperty-edit-label-label');
        $commentLabel = wfMsg('sff-editproperty-edit-comment-label');
        $defaultFormLabel = wfMsg('sff-editproperty-edit-defaultform-label');
        $allowedValuesLabel = wfMsg('sff-editproperty-edit-allowedvalues-label');
        // set 'title' as hidden field, in case there's no URL niceness
        $text =<<<END
    <form action="" method="post">
        <input type="hidden" name="title" value="$specialNamespace:{$this->getLocalName()}" />
        <table>
            <tr VALIGN="top" >
                <td>$nameLabel</td>
                <td>
                    <div id="acProperty">
                        <input id="property_name" name="name" value="{$property->getName()}" />
                        <div id="cntrPropertyName"> </div>
                    </div>
                </td>
                <td><span style="color: red;">$nameError</span></td>
            </tr>
            <tr>
                <td>$typeLabel</td>
                <td><select id="property_dropdown" name="type" onChange="toggleAllowedValues();">\n
END;
        // add blank element to allow user to unset
        $datatypeLabels = array_merge(array(''), $smwContLang->getDatatypeLabels());
        foreach ($datatypeLabels as $label) {
            $text .= str_repeat(' ', 20) . '<option';
            if ($label === $property->getType()) { //@todo will not work in other languages
                $text .= ' selected';
            }
            $text .= ">$label</option>\n";
        }

        $symmetricTrueLabel = wfMsg('sff-editproperty-text-true');
        $symmetricFalseLabel = wfMsg('sff-editproperty-text-false');
        $symmetricNotDefinedLabel = wfMsg('sff-editproperty-text-notdefined');
        $symmetricTrueChecked = $property->getSymmetric() == 'true' ? 'checked' : '';
        $symmetricFalseChecked = $property->getSymmetric() == 'false' ? 'checked' : '';
        $symmetricNotDefinedChecked = $property->getSymmetric() == '' ? 'checked' : '';
        $saveButtonText = wfMsg('savearticle');
        $previewButtonText = wfMsg('preview');
        if ($property->hasArticle()) {
            $cancelLink = $property->getTitle()->getLinkUrl();
        } else {
            $cancelLink = SpecialPage::getPage('SFF_CreateProperty')->getTitle()->getLocalURL();
        }
        $cancelLinkText = wfMsg('cancel');
        $cancelLinkTooltip = wfMsg('sff-cancel-tooltip', $property->getName());
        $text .=<<<END
                </select></td>
            </tr>
            <tr VALIGN="top" >
                <td>$rangeLabel</td>
                <td>
                    <div id="acPropertyRange">
                        <input id="property_range" name="range" value="{$this->buildCommaSeparatedList($property->getRange())}" size="60" />
                        <div id="cntrPropertyRange"> </div>
                    </div>
                </td>
            </tr>
            <tr VALIGN="top" >
                <td>$domainLabel</td>
                <td>
                    <div id="acPropertyDomain">
                        <input id="property_domain" name="domain" value="{$this->buildCommaSeparatedList($property->getDomain())}" size="60" />
                        <div id="cntrPropertyDomain"> </div>
                    </div>
                </td>
            </tr>
            <tr VALIGN="top" >
                <td>$inverseOfLabel</td>
                <td>
                    <div id="acPropertyInverseOf">
                        <input id="property_inverseof" name="inverseof" value="{$this->buildCommaSeparatedList($property->getInverseOf())}" size="60" />
                        <div id="cntrPropertyInverseOf"> </div>
                    </div>
                </td>
            </tr>
            <tr>
                <td>$symmetricLabel</td>
                <td>
                    <input type="radio" name="symmetric" value="true" $symmetricTrueChecked />$symmetricTrueLabel
                    <input type="radio" name="symmetric" value="false" $symmetricFalseChecked />$symmetricFalseLabel
                    <input type="radio" name="symmetric" value="" $symmetricNotDefinedChecked />$symmetricNotDefinedLabel
                </td>
            </tr>
            <tr>
                <td>$labelLabel</td>
                <td>
                    <div id="acPropertyLabel">
                        <input id="property_label" name="label" value="{$property->getLabel()}" />
                        <div id="cntrPropertyLabel"> </div>
                    </div>
                </td>
            </tr>
            <tr>
                <td>$commentLabel</td>
                <td>
                    <div id="acPropertyComment">
                        <input id="property_comment" name="comment" value="{$property->getComment()}" size="60" />
                        <div id="cntrPropertyComment"> </div>
                    </div>
                </td>
            </tr>
            <tr VALIGN="top" >
                <td>$defaultFormLabel</td>
                <td>
                    <div id="acForm">
                        <input id="property_defaultform" name="defaultform" value="{$property->getDefaultForm()}" />
                        <div id="cntrForm"> </div>
                    </div>
                </td>
            </tr>
        </table>
        <div id="allowed_values" style="margin-bottom: 15px;">
            <p>$allowedValuesLabel</p>
            <p><input size="80" name="allowedvalues" value="{$this->listAllowedValues($property)}"></p>
        </div>
        <div class="editButtons">
            <input id="wpSave" type="submit" name="wpSave" value="$saveButtonText" />
            <input id="wpPreview" type="submit" name="wpPreview" value="$previewButtonText" />
            <span class='editHelp'><a href="$cancelLink" title="$cancelLinkTooltip">$cancelLinkText</a></span>
        </div>
    </form>

END;

// Provide autocompletion for field(s):
    $text .= SFFUtils::getCSSText();
    $text .='<script type="text/javascript">/* <![CDATA[ */ ';

// Property range

    $text .= SFFUtils::buildArrayFromNameSpace(SMW_NS_PROPERTY,"localPropList");
    $text .='
var myDS1 = new YAHOO.widget.DS_JSArray (localPropList);
var myAutoComp1 = new YAHOO.widget.AutoComplete("property_name", "cntrPropertyName", myDS1);';

//  Category range
    $text .= SFFUtils::buildArrayFromNameSpace(NS_CATEGORY,"localCatgList");
    $text .='
var myDS2 = new YAHOO.widget.DS_JSArray (localCatgList);
var myAutoComp2 = new YAHOO.widget.AutoComplete("property_range", "cntrPropertyRange", myDS2);

var myAutoComp4 = new YAHOO.widget.AutoComplete("property_domain", "cntrPropertyDomain", myDS2);';

// Default Form
    $text .= SFFUtils::buildArrayFromNameSpace(SF_NS_FORM,"localFormList");
    $text .='
var myDS3 = new YAHOO.widget.DS_JSArray (localFormList);
var myAutoComp3 = new YAHOO.widget.AutoComplete("property_defaultform", "cntrForm", myDS3);';

// Default inverseOf pages:
    $text .= SFFUtils::buildArrayFromNameSpace(SMW_NS_PROPERTY,"localPropList");
    $text .='
var myDS5 = new YAHOO.widget.DS_JSArray (localPropList);
var myAutoComp5 = new YAHOO.widget.AutoComplete("property_inverseof", "cntrPropertyInverseOf", myDS5);';

    $text .='/* ]]> */ </script>';

        $outputPage->addLink( array(
            'rel' => 'stylesheet',
            'type' => 'text/css',
            'media' => "screen, projection",
            'href' => SFFGlobals::getScriptPath() . "/skins/SF_main.css"
            ));
        $outputPage->addScript('<script type="text/javascript">' . $javascript_text . '</script>');
        $outputPage->addHTML($text);
    }


    /**
     * Builds a title for this page, depending on several factors.
     * @param SFFProperty $property the optional property for which the page title needs to be built
     * @return String the page title
     */
    private function buildPageTitle(SFFProperty $property = null) {
        if ($property == null) {
            // no property set, so we are in the choose page
            return wfMsg('sff-editproperty-title-choose');
        }
        $propertyName = $property->getName();
        if ($property->hasArticle()) {
            // existing property article, so we are editing that
            return wfMsg('sff-editproperty-title-edit', $propertyName);
        } else {
            if ($propertyName != '') {
                // new property with name supplied, so we are creating an article for that property
                return wfMsg('sff-editproperty-title-create', $propertyName);
            } else {
                // new property with no name yet, so we are creating some new property
                return wfMsg('sff-editproperty-title-createnew');
            }
        }
    }


    /**
     * Returns a string representation of a string array. It separates the values with the list
     * separator returned by SFFGlobals::getListSeparator(), which would normally be a comma.
     * @param array $array the string array
     * @return String the comma-separated string representation
     */
    private function buildCommaSeparatedList($array) {
        $list = '';
        if ($array) {
            $separator = SFFGlobals::getListSeparator();
            $list = implode($separator . ' ', $array);
        }
        return $list;
    }

    /**
     * Returns a string representation of the allowed values. It separates the values with the list
     * separator returned by SFFGlobals::getListSeparator(), which would normally be a comma. Any
     * occurrences of the separator string in the individual values will be escaped by '\'.
     * @param SFFProperty $property the property whose allowed values will be listed
     * @return String the string representation of the allowed values
     */
    private function listAllowedValues(SFFProperty $property) {
        $list = '';
        $allowedValues = $property->getAllowedValues();
        if ($allowedValues) {
            $separator = SFFGlobals::getListSeparator();
            foreach ($allowedValues as &$value) {
                $value = str_replace($separator, '\\' . $separator, $value);
            }
            $list = implode($separator . ' ', $allowedValues);
        }
        return $list;
    }

    /**
     * Generates the wikitext for a new property.
     * @param SFFProperty $property the property whose wikitext needs to be generated
     * @return String the generated wikitext
     */
    private function generateNewPropertyText(SFFProperty $property) {
        $wikitext = '';

        $type = $property->getType();
        if ($type) {
            $typeLink = $property->getTypeLink();
            $wikitext .= wfMsg('sff-editproperty-text-type', $typeLink);
        }

        $range = $property->getRange();
        if ($range) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $rangeFunctionCall = $property->getRangeFunctionCall();
            $wikitext .= wfMsg('sff-editproperty-text-range', $rangeFunctionCall);
        }

        $domain = $property->getDomain();
        if ($domain) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $domainFunctionCall = $property->getDomainFunctionCall();
            $wikitext .= wfMsg('sff-editproperty-text-domain', $domainFunctionCall);
        }

        $inverseOf = $property->getInverseOf();
        if ($inverseOf) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $inverseOfFunctionCall = $property->getInverseOfFunctionCall();
            $wikitext .= wfMsg('sff-editproperty-text-inverseof', $inverseOfFunctionCall);
        }

        $symmetric = $property->getSymmetric();
        if ($symmetric) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $symmetricLink = $property->getSymmetricLink();
            $wikitext .= wfMsg('sff-editproperty-text-symmetric', $symmetricLink);
        }

        $label = $property->getLabel();
        if ($label) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $labelLink = $property->getLabelLink();
            $wikitext .= wfMsg('sff-editproperty-text-label', $labelLink);
        }

        $comment = $property->getComment();
        if ($comment) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $commentLink = $property->getCommentLink();
            $wikitext .= wfMsg('sff-editproperty-text-comment', $commentLink);
        }

        $defaultForm = $property->getDefaultForm();
        if ($defaultForm) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $defaultFormLink = $property->getDefaultFormLink();
            $wikitext .= wfMsg('sff-editproperty-text-defaultform', $defaultFormLink);
        }

        $allowedValues = $property->getAllowedValues();
        if ($allowedValues) {
            if ($wikitext) {
                $wikitext .= "\n\n";
            }
            $allowsValuesFunctionCall = $property->getAllowsValuesFunctionCall();
            $wikitext .= wfMsgForContent('sff-editproperty-text-allowedvalues', $allowsValuesFunctionCall);
        }

        return $wikitext;
    }

    /**
     * Updates the existing property wikitext with the new values for its meta properties, if any.
     * New meta properties are added, cleared meta properties are removed, and altered meta
     * properties are changed.
     * @param SFFProperty $property the property whose wikitext needs to be updated
     * @return String the updated wikitext
     */
    private function updateExistingPropertyText(SFFProperty $property) {
        $wikitext = $property->getSavedWikiText();

        // update type
        if ($property->isTypeAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-type', $property->getTypeLink());
        } else if ($property->isTypeRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-type', $property->getRawTypeLinkPattern());
        } else if ($property->isTypeChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getTypeLinkPattern(), $property->getTypeLink());
        }

        // update range
        if ($property->isRangeAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-range', $property->getRangeFunctionCall());
        } else if ($property->isRangeRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-range', $property->getRawRangeParserFunctionPattern());
        } else if ($property->isRangeChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getRangeParserFunctionPattern(), $property->getRangeFunctionCall());
        }

        // update domain
        if ($property->isDomainAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-domain', $property->getDomainFunctionCall());
        } else if ($property->isDomainRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-domain', $property->getRawDomainParserFunctionPattern());
        } else if ($property->isDomainChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getDomainParserFunctionPattern(), $property->getDomainFunctionCall());
        }

        // update inverseOf
        if ($property->isInverseOfAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-inverseof', $property->getInverseOfFunctionCall());
        } else if ($property->isInverseOfRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-inverseof', $property->getRawInverseOfParserFunctionPattern());
        } else if ($property->isInverseOfChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getInverseOfParserFunctionPattern(), $property->getInverseOfFunctionCall());
        }

        // update symmetric
        if ($property->isSymmetricAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-symmetric', $property->getSymmetricLink());
        } else if ($property->isSymmetricRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-symmetric', $property->getRawSymmetricLinkPattern());
        } else if ($property->isSymmetricChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getSymmetricLinkPattern(), $property->getSymmetricLink());
        }

        // update label
        if ($property->isLabelAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-label', $property->getLabelLink());
        } else if ($property->isLabelRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-label', $property->getRawLabelLinkPattern());
        } else if ($property->isLabelChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getLabelLinkPattern(), $property->getLabelLink());
        }

        // update comment
        if ($property->isCommentAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-comment', $property->getCommentLink());
        } else if ($property->isCommentRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-comment', $property->getRawCommentLinkPattern());
        } else if ($property->isCommentChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getCommentLinkPattern(), $property->getCommentLink());
        }

        // update default form
        if ($property->isDefaultFormAdded()) {
            $wikitext = $this->addLink($wikitext, 'sff-editproperty-text-defaultform', $property->getDefaultFormLink());
        } else if ($property->isDefaultFormRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-defaultform', $property->getRawDefaultFormLinkPattern());
        } else if ($property->isDefaultFormChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getDefaultFormLinkPattern(), $property->getDefaultFormLink());
        }

        // update allowed values
        if ($property->isAllowedValuesAdded()) {
            $wikitext  = $this->addLink($wikitext, 'sff-editproperty-text-allowedvalues', $property->getAllowsValuesFunctionCall());
        } else if ($property->isAllowedValuesRemoved()) {
            $wikitext = $this->removeLink($wikitext, 'sff-editproperty-text-allowedvalues', $property->getRawAllowsValuesParserFunctionPattern());
        } else if ($property->isAllowedValuesChanged()) {
            $wikitext = $this->changeLink($wikitext, $property->getAllowsValuesParserFunctionPattern(), $property->getAllowsValuesFunctionCall());
        }

        return $wikitext;
    }

    /**
     * Adds a new meta property link to the wikitext.
     * @param String $wikitext the wikitext to which the links needs to be added
     * @param String $textKey the message key for the default sentence for this meta property
     * @param String $link the link to be added
     * @return String the updated wikitext
     */
    private function addLink($wikitext, $textKey, $link) {
        $wikitext .= "\n\n" . wfMsg($textKey, $link) . ' ';
        return $this->trim($wikitext);
    }

    /**
     * Removes an existing meta property link from the wikitext. If the original (default) text as
     * generated by the createNewPropertyText function is still present, it will be completely
     * removed. Otherwise, the link will be replaced by the 'sff-editproperty-text-valuenotset'
     * message. Only the first occurrence will be removed.
     * @param String $wikitext the wikitext that somewhere contains the link to be removed
     * @param String $textKey the message key for the default sentence for this meta property
     * @param String $rawLinkPattern the PCRE pattern that represents the link, without delimiters
     * or modifiers
     * @return String the updated wikitext
     */
    private function removeLink($wikitext, $textKey, $rawLinkPattern) {
        // first, try to completely remove the default sentence added by createNewPropertyText
        // build long pattern including default sentence
        $longPattern = wfMsg($textKey, "\007"); // temporarily insert ASCII bell as link placeholder into sentence to allow escaping PCRE characters
        $longPattern = preg_quote($longPattern, '/'); // now escape all special PCRE characters
        $longPattern = '/' . str_replace("\007", $rawLinkPattern, $longPattern) . '/i'; // replace ASCII bell with link pattern and add '/'
        $wikitext = preg_replace($longPattern, '', $wikitext, 1, $count); // try to remove the long pattern from the text
        if ($count == 0) {
            // could not remove complete sentence, so only replace link with 'not set'
            $replacement = wfMsg('sff-editproperty-text-valuenotset');
            $linkPattern = '/' . $rawLinkPattern . '/i';
            $wikitext = preg_replace($linkPattern, $replacement, $wikitext, 1);
        }
        return $this->trim($wikitext);
    }

    /**
     * Replaces the existing meta property link within the wikitext with the new link. Only the
     * first occurrence will be replaced.
     * @param String $wikitext the wikitext that somewhere contains the link to be changed
     * @param String $linkPattern the PCRE pattern that represents the link
     * @param String $newLink the new link to be inserted
     * @return String the updated wikitext
     */
    private function changeLink($wikitext, $linkPattern, $newLink) {
        // simply replace old link with new (no trimming necessary)
        return preg_replace($linkPattern, $newLink, $wikitext, 1);
    }

    /**
     * Removes leading and trailing whitespace, and consecutive wikitext paragraph separators. Such
     * superfluous whitespace may have been generated by adding or removing meta properties.
     * @param String $wikitext the wikitext to be cleaned up
     * @return String the cleaned up wikitext
     */
    private function trim($wikitext) {
        // remove leading whitespace
        $wikitext = preg_replace('/^\s+/', '', $wikitext);
        // remove trailing whitespace
        $wikitext = preg_replace('/\s+$/', '', $wikitext);
        // collapse consecutive paragraph separators (including whitespace) into a single separator
        $wikitext = preg_replace('/(\n\s*){2,}\n/', "\n\n", $wikitext);
        return $wikitext;
    }

}


/**
 * Class used to build the property query page (used in SFFEditPropery::doChoose).
 */
class SFFPropertyQueryPage extends SMWQueryPage {

    /**
     * Returns the internal name of this QueryPage (should be the same as the special page).
     * @return String the internal name of this QueryPage
     */
    function getName() {
//        return 'editproperty';
        return 'SFF_EditProperty';
    }
    
    /**
     * Returns the page header to be included at the top of the generated page. This function is
     * called by QueryPage::doQuery(). Overrides QueryPage::getPageHeader().
     * @return String the page header (HTML)
     */
    function getPageHeader() {
        // no value for name parameter enables the name field in the form screen
        $path = SpecialPage::getPage('SFF_CreateProperty')->getTitle()->getLocalURL('name=');
        return wfMsg('sff-editproperty-choose-header', $path);
    }

    /**
     * Returns the formatted version for a line from the query result. This function is (indirectly)
     * called by QueryPage::doQuery() for each row in the result set. Overrides
     * QueryPage::formatResult().
     * @param Skin $skin the current MW skin to be used for the formatting
     * @param object $result an object that represents a row from the result set, with the fields as
     * member variables
     * @return mixed a String with the formatted line (HTML), or false if the line should be skipped
     */
    function formatResult($skin, $result) {
        $typestring = '';
        $errors = array();
        wfLoadExtensionMessages('SemanticMediaWiki');
        if ($result[0]->isUserDefined() && ($result[1]<=5)) {
            $errors[] = wfMsg('smw_propertyhardlyused');
        }
        if ($result[0]->isUserDefined() && $result[0]->getWikiPageValue()->getTitle()->exists()) { // FIXME: this bypasses SMWDataValueFactory; ungood
            $types = smwfGetStore()->getPropertyValues($result[0]->getWikiPageValue(), SMWPropertyValue::makeProperty('_TYPE'));
            if (count($types) >= 1) {
                $typestring = current($types)->getLongHTMLText($skin);
            }
            $proplink = $skin->makeKnownLinkObj(SpecialPage::getPage('SFF_EditProperty')->getTitle(), str_replace('_', ' ', $result[0]->getDBkey()), 'name=' . $result[0]->getDBkey());
        } elseif ($result[0]->isUserDefined()) {
            $errors[] = wfMsg('smw_propertylackspage');
            $proplink = $skin->makeBrokenLinkObj(SpecialPage::getPage('SFF_EditProperty')->getTitle(), str_replace('_', ' ', $result[0]->getDBkey()), 'name=' . $result[0]->getDBkey());
        } else { // predefined property
            $type = $result[0]->getTypesValue();
            $typestring = $type->getLongHTMLText($skin);
            if ($typestring == '') $typestring = '&ndash;'; /// FIXME some types of builtin props have no name, and another message should be used then
            $title = SpecialPage::getPage('SFF_EditProperty')->getTitle();
            if ($result[0]->getWikiPageValue()->getTitle()->exists()) {
                $proplink = $skin->makeKnownLinkObj($title, str_replace('_', ' ', $result[0]->getDBkey()), 'name=' . $result[0]->getDBkey());
            } else {
                $proplink = $skin->makeBrokenLinkObj($title, str_replace('_', ' ', $result[0]->getDBkey()), 'name=' . $result[0]->getDBkey());
            }
            $proplink = $this->highlightText($proplink);
        }
        if ($typestring == '') {
            $type = SMWDataValueFactory::newPropertyObjectValue(SMWPropertyValue::makeProperty('_TYPE'));
            $type->setDBkeys(array('_wpg'));
            $typestring = $type->getLongHTMLText($skin);
            if ($result[0]->getWikiPageValue()->getTitle()->exists()) { // print only when we did not print a "nopage" warning yet
                $errors[] = wfMsg('smw_propertylackstype', $type->getLongHTMLText());
            }
        }
        return wfMsg('smw_property_template', $proplink, $typestring, $result[1]) . ' ' . smwfEncodeMessages($errors);
    }

    private function highlightText($text) {
        SMWOutputs::requireHeadItem(SMW_HEADER_TOOLTIP);
        wfLoadExtensionMessages('SemanticMediaWiki');
        return '<span class="smwttinline"><span class="smwbuiltin">' . $text .
                '</span><span class="smwttcontent">' . wfMsgForContent('smw_isspecprop') . '</span></span>';
     }


    /**
     * Returns the (partial) 'list' of used and/or defined properties to be displayed. If the
     * $requestOptions parameter contains limit or offset fields with positive values, then these
     * will be taken into account. Otherwise, all properties will be returned.
     * @param SMWRequestOptions $requestOptions the options to be used for the result
     * @return array(array(SMWPropertyValue, Integer)), which is an array of tuples of
     * (SMWPropertyValue, Integer), where the Integer value indicates the number of times that the
     * property has been used in the wiki.
     */
	function getResults($requestOptions) {
        // store requested limit and offset
        $requestedLimit = $requestOptions->limit;
        $requestedOffset = $requestOptions->offset;
        // reset to 0 to request all (required for merging and sorting)
        $requestOptions->limit = 0;
        $requestOptions->offset = 0;
        // returns array of tuples (array) of (SMWPropertyValue, Integer)
        $usedProperties = smwfGetStore()->getPropertiesSpecial($requestOptions);
        // returns array of SMWPropertyValue
        $unusedProperties = smwfGetStore()->getUnusedPropertiesSpecial($requestOptions);
        $properties = array();
        foreach ($usedProperties as $propertyInfo) {
            $properties[$propertyInfo[0]->getDBkey()] = $propertyInfo;
        }
        foreach ($unusedProperties as $property) {
            $properties[$property->getDBkey()] = array($property, 0);
        }
        if ($requestOptions->sort) {
            if ($requestOptions->ascending) {
                ksort($properties);
            } else {
                krsort($properties);
            }
        }
        $properties = array_values($properties);
        if ($requestedLimit <= 0) {
            return $properties;
        } else {
            if ($requestedOffset < 0) {
                $requestedOffset = 0;
            }
            return array_slice($properties, $requestedOffset, $requestedLimit);
        }
    }
    
}
?>