<?php
/**
 * A special page holding a form that allows the user to create a template
 * with semantic fields.
 *
 * @author Yaron Koren
 * @author Henri Gerrits
 */

/**
 * Protect against register_globals vulnerabilities.
 * This line must be present before any global variable is referenced.
 */
if (!defined('MEDIAWIKI')) die();

class SFFEditTemplate extends SFFTemplateSpecialPage {

    /**
     * Constructs the special page and registers it with MW.
     */
    function __construct() {
        SpecialPage::SpecialPage('SFF_EditTemplate');
    }

}

class SFFCreateTemplate extends SFFTemplateSpecialPage {

    /**
     * Constructs the special page and registers it with MW.
     */
    function __construct() {
        SpecialPage::SpecialPage('SFF_CreateTemplate');
    }

}

/**
 * A special page providing a form that allows the user to create or edit a template.
 *
 * @author Henri Gerrits
 */
abstract class SFFTemplateSpecialPage extends SpecialPage {

    /**
     * Class entry point that displays the appropriate page depending on the circumstances.
     * Overrides SpecialPage::execute().
     * @param <type> $query not used
     */
    function execute($query) {
        wfLoadExtensionMessages('SemanticForms');
        wfLoadExtensionMessages('SFF_extension');
        $this->setHeaders();
        $request = SFFGlobals::getRequest();
        $template = new SFFTemplate($request);
        if (!$request->getCheck('name')) {
            $this->doChoose();
        } else {
            $isSavePage = $request->getCheck('wpSave');
            $isPreviewPage = $request->getCheck('wpPreview');
            if ($isSavePage || $isPreviewPage) {
                $this->doView($template, $isSavePage, $isPreviewPage);
            } else {
                $this->doEdit($template);
            }
        }
    }

    /**
     * Builds a 'choose' page, containing an option to create a new template, as well as a list of
     * existing template that can be edited.
     */
    private function doChoose() {
        $outputPage = SFFGlobals::getOutputPage();
        $outputPage->setPageTitle($this->buildPageTitle());

        list($limit, $offset) = wfCheckLimits();
        $queryPage = new SFFTemplateQueryPage();
        $queryPage->doQuery($offset, $limit);
        SMWOutputs::commitToOutputPage($outputPage); // make sure locally collected output data is pushed to the output!
    }


    /**
     * Builds the 'view' page after the save or preview buttons have been pressed.
     * @param SFFTemplate $template the template to be viewed on the page
     * @param Boolean $isSavePage true if save button has been pressed previously
     * @param Boolean $isPreviewPage true if preview button has been pressed previously
     */
    private function doView(SFFTemplate $template, $isSavePage, $isPreviewPage) {
        if ($template->getName() == '') {
            $this->doEdit($template, wfMsg('sff-edittemplate-edit-error-nameblank'));
            return;
        }
        // @todo test whether name for new template already exists
        $outputPage = SFFGlobals::getOutputPage();
        $outputPage->setPageTitle($this->buildPageTitle($template));

        // redirect to wiki interface
        $outputPage->setArticleBodyOnly(true);
        $title = $template->getTitle();
        $fullText = SFFTemplateField::createTemplateText($template->getName(), $template->getFields(), $template->getCategory(), $template->getAggregatingProperty(), $template->getAggregationLabel(), $template->getTemplateFormat());
        $fullText = str_replace('"', '&quot;', $fullText);
        $text = SFUtils::printRedirectForm($title, $fullText, '', $isSavePage, $isPreviewPage, false, false, false, null, null);
        $outputPage->addHTML($text);
    }


    /**
     * Builds the edit template with form page and places it in the global OutputPage object.
     * @param SFFTemplate $template the template to be edited in the form
     * @param String $nameError any errors to be displayed related to the name field
     */
    private function doEdit(SFFTemplate $template, $nameError = '') {

        $outputPage = SFFGlobals::getOutputPage();
        $mwContLang = SFFGlobals::getMWContLang();
        $user = SFFGlobals::getUser();

        wfLoadExtensionMessages('SemanticForms');

        $outputPage->setPageTitle($this->buildPageTitle($template));

        $allProperties = self::getAllPropertyNames();

        $name = $template->getName();
        $category = $template->getCategory();
        $currentId = 1;
        $fields = $template->getFields();
        $aggregatingProperty = $template->getAggregatingProperty();
        $aggregationLabel = $template->getAggregationLabel();
        $templateFormat = $template->getTemplateFormat();

        $text = '	<form action="" method="post">' . "\n";
        // set 'title' field, in case there's no URL niceness
        $namespaceLabels = $mwContLang->getNamespaces();
        $specialNamespace = $namespaceLabels[NS_SPECIAL];
        $text .= '    <input type="hidden" name="title" value="' . $specialNamespace . ':SFF_EditTemplate">' . "\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_namelabel') .
'<div id="acTemplate">
    <input size="25" id="template_name" name="name" value="' . $name . '">
    <div id="cntrTemplate"> </div>
</div>
<span style="color: red;">' . $nameError . "</span></p>\n";

 $text .= '	<p>' . wfMsg('sf_createtemplate_categorylabel') .
'<div id="acCategory">
    <input size="25" id="category" name="category" value="' . $category . '">
    <div id="cntrCategory"> </div>
</div>
</p>' . "\n";
        $text .= "	<fieldset>\n";
        $text .= '	<legend>' . wfMsg('sf_createtemplate_templatefields') . "</legend>\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_fieldsdesc') . "</p>\n";

        foreach ((array)$fields as $i => $field) {
            $text .= $this->printFieldEntryBox($i + 1, $field, $allProperties);
        }
        $newField = new SFFTemplateField();
        $text .= $this->printFieldEntryBox('new', $newField, $allProperties);
            
        $text .= '	<p><input type="submit" value="' . wfMsg('sf_createtemplate_addfield') . '"> ';
        $text .= '<input type="submit" id="loadproperties" name="loadproperties" value="' . wfMsg('sff-edittemplate-edit-loadproperties') . '" /></p>' . "\n";
        $text .= "	</fieldset>\n";
        $text .= "	<fieldset>\n";
        $text .= '	<legend>' . wfMsg('sf_createtemplate_aggregation') . "</legend>\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_aggregationdesc') . "</p>\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_semanticproperty') . " " . $this->printPropertiesDropdown($allProperties, "aggregation", $aggregatingProperty). "</p>\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_aggregationlabel') . ' <input size="25" name="aggregationlabel" value="' . $aggregationLabel . '"></p>' . "\n";
        $text .= "	</fieldset>\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_outputformat') . "\n";
        $standardChecked = ($template->getTemplateFormat() == 'standard' ? ' checked ' : ' ');
        $infoboxChecked = ($template->getTemplateFormat() == 'infobox' ? ' checked ' : ' ');
        $text .= '	<input type="radio" name="templateformat"' . $standardChecked . 'value="standard">' . wfMsg('sf_createtemplate_standardformat') . "\n";
        $text .= '	<input type="radio" name="templateformat"' . $infoboxChecked . 'value="infobox">' . wfMsg('sf_createtemplate_infoboxformat') . "</p>\n";
    	$saveButtonText = wfMsg('savearticle');
        $previewButtonText = wfMsg('preview');
        if ($template->hasArticle()) {
            $cancelLink = $template->getTitle()->getLinkUrl();
        } else {
            $cancelLink = SpecialPage::getPage('SFF_CreateTemplate')->getTitle()->getLocalURL();
        }
        $cancelLinkText = wfMsg('cancel');
        $cancelLinkTooltip = wfMsg('sff-cancel-tooltip', $template->getName());
        $text .=<<<END
        <div class="editButtons">
            <input id="wpSave" type="submit" name="wpSave" value="$saveButtonText">
            <input id="wpPreview" type="submit" name="wpPreview" value="$previewButtonText">
            <span class='editHelp'><a href="$cancelLink" title="$cancelLinkTooltip">$cancelLinkText</a></span>
        </div>
	</form>

END;
        $skin = $user->getSkin();
        $editPropertySpecialPage = SpecialPage::getPage('SFF_EditProperty');
        $editPropertyLink = $skin->makeKnownLinkObj($editPropertySpecialPage->getTitle(), $editPropertySpecialPage->getDescription());
        $text .= "	<br /><hr /><br />\n";
        $text .= "	<p>$editPropertyLink.</p>\n";

        $createCategorySpecialPage = SpecialPage::getPage('CreateCategory');
        $createCategoryLink = $skin->makeKnownLinkObj($createCategorySpecialPage->getTitle(), $createCategorySpecialPage->getDescription());
        $text .= "	<br /><hr /><br />\n";
        $text .= "	<p>$createCategoryLink.</p>\n";

// Provide autocompletion for field(s):
        $text .= SFFUtils::getCSSText();
        $text .='<script type="text/javascript">/* <![CDATA[ */ ';

// Template

        $text .= SFFUtils::buildArrayFromNameSpace(NS_TEMPLATE, "localPropList");
        $text .='var myDS1 = new YAHOO.widget.DS_JSArray (localPropList);
var myAutoComp1 = new YAHOO.widget.AutoComplete("template_name", "cntrTemplate", myDS1);';

//  Category
        $text .= SFFUtils::buildArrayFromNameSpace(NS_CATEGORY, "localCatgList");
        $text .='var myDS2 = new YAHOO.widget.DS_JSArray (localCatgList);
var myAutoComp2 = new YAHOO.widget.AutoComplete("category", "cntrCategory", myDS2);';


        $text .='/* ]]> */ </script>';
        $outputPage->addLink(array(
            'rel' => 'stylesheet',
            'type' => 'text/css',
            'media' => 'screen, projection',
            'href' => SFFGlobals::getScriptPath() . '/skins/SF_main.css'
        ));
        $outputPage->addHTML($text);
    }

    /**
     * Builds a title for this page, depending on several factors.
     * @param SFFTemplate $template the optional template for which the page title needs to be built
     * @return String the page title
     */
    private function buildPageTitle(SFFTemplate $template = null) {
        if ($template == null) {
            // no template set, so we are in the choose page
            return wfMsg('sff-edittemplate-title-choose');
        }
        $templateName = $template->getName();
        if ($template->hasArticle()) {
            // existing template article, so we are editing that
            return wfMsg('sff-edittemplate-title-edit', $templateName);
        } else {
            // no article yet, so we are creating some new template
            return wfMsg('sff-edittemplate-title-create');
        }
    }

    /**
     * Returns a sorted array of the names of all properties used and/or defined in the wiki.
     * @return array(String) the property names
     */
    private static function getAllPropertyNames() {
        $allProperties = array();

        // set limit on results - a temporary fix until SMW's
        // getProperties() functions stop requiring a limit
        $options = new SMWRequestOptions();
        $options->limit = 10000;
        $usedProperties = smwfGetStore()->getPropertiesSpecial($options);
        foreach ($usedProperties as $property) {
            $propertyName = SFUtils::getPropertyName($property[0]);
            // @TODO: Hack. Sometimes getPropertiesSpecial returns SFF's special properties twice.
            // Don't know why; possibly because special properties were made special AFTER using
            // them in the development wiki.
            // This only adds property names that are not yet in the list.
            if (!in_array($propertyName, $allProperties)) {
                $allProperties[] = $propertyName;
            }
        }
        $unusedProperties = smwfGetStore()->getUnusedPropertiesSpecial($options);
        foreach ($unusedProperties as $property) {
            $allProperties[] = SFUtils::getPropertyName($property);
        }

        // sort properties list alphabetically
        sort($allProperties);
        return $allProperties;
    }

    /**
     * Generates the HTML code for the property dropdown box.
     * @param array(String) $allProperties an array of used and/or defined property names
     * @param String $id the id suffix to be added to the select element's name attribute
     * @param String $selectedProperty the name of the currently selected property name
     * @return String the HTML code for the property dropdown box
     */
    private function printPropertiesDropdown($allProperties, $id, $selectedProperty) {
        $dropdownHtml = "<select name=\"semanticproperty_$id\">\n";
        $dropdownHtml .= "<option value=\"\"></option>\n";
        foreach ($allProperties as $propertyName) {
            $selected = ($selectedProperty == $propertyName) ? 'selected' : '';
            $dropdownHtml .= "<option value=\"$propertyName\" $selected>$propertyName</option>\n";
        }
        $dropdownHtml .= "</select>\n";
        return $dropdownHtml;
    }

    /**
     * Generates the HTML code for the template field entry box, which is displayed as a group of
     * input elements.
     * @param String $id the id suffix to be added to the elements' name attribute
     * @param SFFTemplateField $field the template field to be displayed
     * @param array(String) $allProperties an array of used and/or defined property names
     * @return String the HTML code for the template field entry box
     */
    private function printFieldEntryBox($id, $field, $allProperties) {
        wfLoadExtensionMessages('SemanticForms');
        $dropdownHtml = $this->printPropertiesDropdown($allProperties, $id, $field->semantic_property);
        $text = '	<div class="fieldBox">' . "\n";
        $text .= '	<p>' . wfMsg('sf_createtemplate_fieldname') . ' <input size="15" name="name_' . $id . '" value="' . $field->field_name . '">' . "\n";
        $text .= '	' . wfMsg('sf_createtemplate_displaylabel') . ' <input size="15" name="label_' . $id . '" value="' . $field->label . '">' . "\n";
        $text .= '	' . wfMsg('sf_createtemplate_semanticproperty') . ' ' . $dropdownHtml . "</p>\n";
        $checked_str = ($field->is_list) ? " checked" : "";
        $text .= '	<p><input type="checkbox" name="is_list_' . $id . '"' .  $checked_str . '> ' . wfMsg('sf_createtemplate_fieldislist') . "\n";

        if ($id != 'new') {
            $text .= '	&nbsp;&nbsp;<input name="del_' . $id . '" type="submit" value="' . wfMsg('sf_createtemplate_deletefield') . '">' . "\n";
        }
        $text .= <<<END
</p>
</div>

END;
        return $text;
    }
}


/**
 * Class used to build the template query page (used in SFFEditTemplate::doChoose).
 */
class SFFTemplateQueryPage extends QueryPage {

    /**
     * Returns the internal name of this QueryPage (should be the same as the special page).
     * @return String the internal name of this QueryPage
     */
    function getName() {
        return 'SFF_EditTemplate';
    }

    /**
     * Returns whether this is an expensive operation or not.
     * @return true
     */
    function isExpensive() { return true; }

    /**
     * Returns the page header to be included at the top of the generated page. This function is
     * called by QueryPage::doQuery(). Overrides QueryPage::getPageHeader().
     * @return String the page header (HTML)
     */
    function getPageHeader() {
        // no value for name parameter enables the name field in the form screen
        $path = SpecialPage::getPage('SFF_CreateTemplate')->getTitle()->getLocalURL('name=');
        return wfMsg('sff-edittemplate-choose-header', $path);
    }

    /**
     * Returns the SQL select query string to be used for generating the result set.
     * @return String the SQL select query string
     */
    function getSQL() {
        $templateNamespace = NS_TEMPLATE;
        $dbr = wfGetDB(DB_SLAVE);
        $page = $dbr->tableName('page');
        // QueryPage uses the value from this SQL in an ORDER clause,
        // so return page_title as title.
        return "SELECT 'Templates' as type,
            page_title as title,
            page_title as value
            FROM $page
            WHERE page_namespace = $templateNamespace";
    }

    /**
     * Returns whether the result should sorted in descending order.
     * @return false
     */
    function sortDescending() {
        return false;
    }

    /**
     * Returns the category name that this template refers to.
     * @param Article $templateArticle the article for the template
     * @return String the template name
     */
    function getCategoryDefinedByTemplate($templateArticle) {
        $templateText = $templateArticle->getContent();
        if (preg_match_all('/\[\[Category:([\w ]*)/', $templateText, $matches)) {
            // get the last match - if there's more than one
            // category tag, there's a good chance that the last
            // one will be the relevant one - the others are
            // probably part of inline queries
            return trim(end($matches[1]));
        }
        return '';
    }


    /**
     * Returns the formatted version for a line from the query result. This function is (indirectly)
     * called by QueryPage::doQuery() for each row in the result set. Overrides
     * QueryPage::formatResult().
     * @param Skin $skin the current MW skin to be used for the formatting
     * @param object $result an object that represents a row from the result set, with the fields as
     * member variables
     * @return mixed a String with the formatted line (HTML), or false if the line should be skipped
     */
    function formatResult($skin, $result) {
        wfLoadExtensionMessages('SemanticForms');
        $title = Title::makeTitle( NS_TEMPLATE, $result->value );
        $text = $skin->makeLinkObj(SpecialPage::getPage('SFF_EditTemplate')->getTitle(), str_replace('_', ' ', $result->value), 'name=' . $result->value);
        $category = $this->getCategoryDefinedByTemplate(new Article($title));
        if ($category != '') {
            $text .= ' ' . wfMsg('sf_templates_definescat') . ' ' . SFUtils::linkText(NS_CATEGORY, $category);
        }
        return $text;
    }

}
?>
